//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

///-*- c++ -*-//////////////////////////////////////////////////////////
//
// FColorModel.h - contains all colors for GUI
//
// $Id: FColors.h,v 1.3 2003/03/12 13:35:05 aarn Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef FCOLORMODEL_H
#define FCOLORMODEL_H

#include <qcolor.h>
#include <qpixmap.h>
#include <qpainter.h>

#include <vector>
#include <string>

#include "Exception.h"

class FColors
{
public:

  class ParamError : public Exception
  {
  public:
    ParamError(const char* name) 
      : Exception( std::string("(ParamError: ")+
		   std::string( name )+ std::string(")")) 
    {};
  };

  /** store frequently used values (NB: call Config::readFile before !!!)
   */
  static void init() throw (ParamError);
  
  // commonly used constants as class members
  // fill colors
  static QBrush leftTeam;
  static QBrush leftGoalie;
  static QBrush rightTeam;
  static QBrush rightGoalie;

  // pen colors
  static QPen   playerNorm;
  static QPen   playerKick;
  static QPen   playerColl;
  static QPen   goalieCatch;
  static QPen   playerTackDone;
  static QPen   playerTackFault;

  // ball colors 
  static QBrush ballFill;
  static QPen   ballPen;

  static QColor background;
  static QBrush field;
  static QPen   lines;
  static QColor goal;

  // colors used for viewcone (are added to background!)
  static QBrush viewFeel;
  static QBrush viewNear;
  static QBrush viewFar;

  // color for markers
  static QPen   marker;

  // color for neutral log file entries
  static QColor dflLogColor;

  /** get number of user defined agent drawing colors 
   */
  static int numAgentColors();
  /** get agent drawing color (for "draw-macros")
   */
  static const QColor & agentColor(int i);
  static QPixmap *createColorIcon(const int color);
  static QPixmap *createTeamIcon(const bool left);

  /** parameters to onfigure behaviour of FColorSelects
   */
  static bool actLog;
  static bool actWM;
  static bool actDraw;
  static bool autoAct;

  class Flags {
  public:
    bool active;
    bool leftTeam;
    int  playerNo;
    bool drawTraces;
    bool WM;
    bool log;
    int  autoExpLevel;
    Flags() {};
    void   readParam(const char *);
  };

  static const Flags & agentFlags(int i);

private:
  static QColor readParam(const char *) throw (ParamError);
  static int    readParamAsInt(const char *) throw (ParamError);
  static bool   readParamAsBool(const char *) throw (ParamError);
  static std::vector<QColor>agentC;
  static std::vector<Flags>agentF;
};

#endif
