//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FLogWindow.h - Implements a simple logfile window using a QListView
//
// $Id: FLogWindow.h,v 1.1.1.1 2002/10/08 12:40:43 cm Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef FLOGWINDOW_H
#define FLOGWINDOW_H

#include <qlistview.h>
#include <qpopupmenu.h>
#include <qpixmap.h>
#include "FColors.h"
#include "Config.h"

#if (QT_VERSION >= 200)
#define QT2
#endif

#include <qtooltip.h>
#include "Trace.h"

class FLogWindow : public QListView
{
  Q_OBJECT

public:
  FLogWindow(QWidget * parent = 0, const char * name = 0);
  ~FLogWindow();
  
  // returns the state of the "clear before new step" flag
  bool clearStep()
  { return contextMenu->isItemChecked(1); }

  bool autoExpand()
  { return contextMenu->isItemChecked(2); }

public slots:
  void addLine( const int time, const int color, const int plNum,
		const char *msg);
  void addLine( const int time, const int color, const int plNum,
		const int module, const int level,
		const char *file, const char *line, const char *func,
		const char *msg);
  void addLine( const int time, const char *msg)
  { addLine( time, -1, -1, msg ); }

  void clearList() 
  { clear(); 
    for (int i = 0; i < 13; ++i) {
      players[i] = 0;
      for (int j = 0; j < Trace::NUM_MODULES; ++j)
	modules[i][j] = 0;
    }
  }

  void toggleClearMode()
  { contextMenu->setItemChecked(1, !clearStep()); }

  void toggleExpMode()
  { contextMenu->setItemChecked(2, !autoExpand()); }

  void autoExpLevelChanged( const int num, const int level )
  { 
    if (0 <= num && num < FColors::numAgentColors()) {
      autoExpLevel[num]=level;
    }
  };

protected slots:

  void sigExpanded( QListViewItem * item );

signals:
  void closedWindow();

protected:
  // some strange renaming in QT2 ...
#ifdef QT2
  virtual void contentsMouseReleaseEvent ( QMouseEvent * );
#else
  virtual void mouseReleaseEvent ( QMouseEvent * );
#endif

  virtual void closeEvent( QCloseEvent * e )
  { e->accept(); emit closedWindow(); }
  
private:
  QPopupMenu *contextMenu;
  QPixmap    **entryIcons;
  const int  maxEntryIcons;
  QListViewItem * modules[13][Trace::NUM_MODULES];  // 12 players + 1 coach
  QListViewItem * players[13];
  int *autoExpLevel;

  class FLogToolTip : public QToolTip {
  public:
    FLogToolTip( QListView * parent ) : QToolTip(parent) {};
  protected:
    void maybeTip( const QPoint & );
  };
  FLogToolTip * tt;

  class FLogItem : public QListViewItem {
  public:
    FLogItem( int level, QListViewItem * parent = 0,
	      QListViewItem * after = 0)
      : QListViewItem(parent, after), level(level) {};
    int level;
    QString file, line, function;
  };

  QListViewItem * lastLVI;
};

#endif
