//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "FunsselIO.h"
#include "Base.h"
#include <constants.h>
#include <cstdio>
#include <string>

#include "FunsselStrings.h"

using namespace std;

//#define DEBUG

// cwz-build-command: make FUNSSELMon

FunsselIO::FunsselIO(int port, const char *host) :
  curtime(-1),
  smsg(SoccerMessage::emptyMessage),
  sock(port, host, MAX_SERVER_MSG_SIZE)
{
}

void FunsselIO::fetch(int w)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (fetch %d))", w);
  sendMsg(buf);
}

void FunsselIO::fetchRel(int w)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (fetch %c%d))", (w > 0) ? 'f' : 'b', abs(w));
  sendMsg(buf);
}

void FunsselIO::maxRunTime(int t)
{
  char buf[64];
  sprintf(buf, FILTER_STRING " (max_run %d))", t);
  sendMsg(buf);
}

void FunsselIO::startServer()
{
  sendMsg(FILTER_STRING " (server start))");
}

void FunsselIO::stopServer()
{
  sendMsg(FILTER_STRING " (server stop))");
}

void FunsselIO::filterOn()
{
  sendMsg(FILTER_STRING " (filter on))");
}

void FunsselIO::filterOff()
{
  sendMsg(FILTER_STRING " (filter off))");
}

void FunsselIO::quit()
{
  sendMsg(FILTER_STRING " (quit))");
}

void FunsselIO::init()
{
  sendMsg(FILTER_STRING " (dispinit))");
}

void FunsselIO::syncPlay()
{
  sendMsg(FILTER_STRING " (sync_mode))");
}

void FunsselIO::sendMsg(const string &s)
{
#ifdef DEBUG
  if (s.length() < 64)
    cerr << "got msg: " << s << endl;
  else
    cerr << "got msg: long crap of size " << s.length() << endl;
#endif

  sock.send(s);
}

FunsselIO::Result FunsselIO::getNextMsg()
{
  string str = sock.poll();
  if (str == string())
    return NoMsg;

#ifdef DEBUG
  if (str.length() < 64)
    cerr << "got msg: " << str << endl;
  else
    cerr << "got msg: long crap of size " << str.length() << endl;
#endif

  if (!strncmp(str.data(), FILTER_STRING, FILTER_STRING_LEN)) {
    // ok, a filtered msg. Depack
    char *parse;
    srvtime = strtol(str.data() + FILTER_STRING_LEN + 1, &parse, 10);
    parse++;
    if (!strncmp(parse, "(error", 6)) {
      cerr << "got error from FUNSSEL: (" <<  parse << endl;
      return NoMsg;
    }
    if (!strncmp(parse, "(server time changed)", 21))
      return NewSrvTime;
    if (!strncmp(parse, "(dispeol)", 9))
      return MsgEol;
    if (!strncmp(parse, "(insert failed)", 15))
      return MsgDropped;
    if (!strncmp(parse, "(max_run ok)", 12))
      return NoMsg;
    if (!strncmp(parse, "(welcome)", 9))
      return NoMsg;
    if (!strncmp(parse, "(server started)", 15))
      return ServerStarted;
    if (!strncmp(parse, "(databurst ", 11) ||
	!strncmp(parse, "(datastream", 11)) {
      parse += 11;
      if (sscanf(parse, "%d", &curtime) != 1) {
	cerr << "cannot parse data burst header " << parse << endl;
	return NoMsg;
      }
      return NewDataburst;
    }
    if (!strncmp(parse, "(dispmsg", 8)) {
      int scanned, len;
      int i_sender, i_recver;
      parse += 9;
      if (sscanf(parse, "%d %d %d %d %d %n", &smsg.time,
		 &smsg.durtn, &i_sender, &i_recver,
		 &len, &scanned) != 5) {
	cerr << "cannot parse dispmsg " << parse << endl;
	return NoMsg;
      }

      smsg.sender = (SoccerActor)i_sender;
      smsg.recver = (SoccerActor)i_recver;
      smsg.message.assign(parse + scanned, len);
      return MsgReady;
    }
  }
  // ignore everything else
  // but not quietly
  cerr << "FUNSSEL sent unwanted " << str << endl;
  return NoMsg;
}


