//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef ICFILE_H
#define ICFILE_H
#include <vector>
#include <string>
#include "Exception.h"

class ICFile {
  std::string name;
  void *file;
  bool  wopen;

  std::vector<int> record_starts;
public:
  ICFile() : file(NULL) {};
  ICFile(const std::string &name) throw (Exception);
  ~ICFile();

  bool open(const std::string &name, bool write = false) throw (Exception);

  void close();

  void flush();

  bool isWritable()       const { return wopen; };

  void   push_back(const std::string &data)   throw (Exception);
  /** read a record.
      @param record  the record to read
      @param size    the minimum size to read. If less bytes are
                     available for that record, less bytes are returned.
		     A value of -1 returns the complete record.
  */
  std::string read(int record, int size = -1) throw (Exception);

  int    getNumRecords() const { return record_starts.size(); }
  std::string getFileName()   const { return name; }
private:
  void setupIndexList()  throw (Exception);
  int  readNoSeek(std::string &res, int size);
};

#endif







