//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "onlinecoach.h"
#include "coordinates.h"
#include "tacticalmodule.h"
#include "PositionModule.h"
#include "passmodule.h"
#include "goalshotmodule.h"
#include "Team.h"

struct DesiredType{
    double decay_inertia_delta;
    double decay_inertia_weight;
    double kickmargin_rand_delta;
    double kickmargin_rand_weight;
    double dpr_stamina_inc_max_delta;
    double dpr_stamina_inc_max_weight;
    double extra_stamina_effort_delta;
    double extra_stamina_effort_weight;
};

class MrbCoach : public Onlinecoach
{
public:
  MrbCoach( const char *team, const char *host, const int port,
	    const char *configfile, const char* traceCfg,
	    const bool FUNSSELmode = false );
  ~MrbCoach();

  virtual void act();
  virtual void analyze();

private:
  bool sayMessageNow;
  play_mode lastPlayMode;

  int  numMsgsLeft;          // number of audio messages left

  long lastNewMessageTime;   // last time message counter was reset
  
  long lastAumsgTime;        // last time coach sent an audio message
  
  TacticalModule * tactical;

  PassEval       * passEval;
  GoalShotEval   * goalShotEval;
  PositionEval   * positionEval;
  Team           * otherTeam;

  // compares values in .rcssserver-player.conf with the expected values 
  bool okPlayerConf(); 

  void chooseHeterogenousPlayers();
  void readAgentConf();
  void choosePlayerType(const int player_nr);
  double calcUtility(const int player_nr,int player_type);
  
  DesiredType desiredType[12];

  int calcPlayerForKickUs() const;

  void doSub( const int player, const int type );
  void initSubs();
  void printStats();

  // data for subs module:
  // number of subs left in Game (after KICK_OFF !)
  // set to 0 to avoid subs in online game
  unsigned int numSubs;
  //  static unsigned int numSubs(0);
  bool initDone;
  
  // information about how many players of one type are left
  int numPlTypes;
  int *numPlType;
  
  // types for given players
  int plType[12];

}; // class MrbCoach
