#include <iostream>
#include <fstream>
#include <strstream>
#include <netinet/in.h>
#include <string>
#include <SServer.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include <MsgStore.h>

#include "ballCorrelation.h"

// cwz-build-command: gmake

const int grid = 10000;
const double cutoff_dist = 100;

const int n_teams = 10;
const char *teams[n_teams] = {
  "11MonkeysZ00",
  "ATH2000neu",
  "BS2001",
  "Cyberoos01",
  "FCPortugal00",
  "MRB01",
  "Robolog01",
  "TutGroove01",
  "VirtualWerder01",
  "YowAI00" 
};

bool attached;
string team[2];
int    team_id[2] = { 0, 0 };

// ball distance statistics
int ball_dist[n_teams][grid];


void tryAttachTeams(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return;

  team_id[0] = team_id[1] = -1;

  if (show2->team[0].name[0]) {
    team[0] = show2->team[0].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[0] == teams[t] ||
	  team[0] == string(teams[t]) + "-2")
	team_id[0] = t;
    }
  }
  if (show2->team[1].name[0]) {
    team[1] = show2->team[1].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[1] == teams[t] ||
	  team[1] == string(teams[t]) + "-2")
	team_id[1] = t;
    }
  }
}

int main(int argc, const char **argv)
{
  try {
    for (int arg = 1; arg < argc; arg++) {
      cerr << 100*arg/argc << "% done" << endl;
      MsgStore store(argv[arg]);
      attached = false;

      for (int time = 0;; time++) {
	if (!store.fetchTime(time))
	  break;

	if ((time % 100) == 0)
	  cerr << "time " << time << "\r";

	SoccerMessage msg;
	while ((msg = store.getNextMessage()).isValid()) {
	  if (!attached) {
	    tryAttachTeams(msg);
	    if (team_id[0] != -1 && team_id[1] != -1)
	      attached = 1;
	  }

	  if (attached) {
	    SServer::showinfo_t2 *show2 = getShowInfo(msg);
	    if (!show2)
	      continue;

	    Vector2d bpos = getBallPos(show2);
	    for (int i = 0; i < 22; i++) {
	      Vector2d ppos = getPlayerPos(show2, i);
	      double dist2 = sqrnorm(ppos - bpos);
	      if (dist2 < sqr(cutoff_dist)) {
		double dist = sqrt(dist2);
		int g = int(floor(dist/cutoff_dist*grid));
		if (g >= grid)
		  g = grid - 1;
		int team = (i < 11) ? 0 : 1;
		ball_dist[team_id[team]][g]++;
	      }
	    }
	    // finish this time step since we had a showinfo
	    break;
	  }
	}
      }
    }

    for (int team = 0; team < n_teams; team++) {
      int total = 0;
      for (int g = 0; g < grid; g++)
	total += ball_dist[team][g];

      // we do not have data from that team
      // hope that there is old data
      if (total == 0)
	continue;

      strstream name;
      name << "b_dist_0.01_" << teams[team] << (char) 0;

      ofstream out(name.str());

      for (int g = 0; g < grid; g++) {
	double d = cutoff_dist*double(g)/grid;
	out << d << " " << double(ball_dist[team][g])/total << endl;
      }
    }

    // calculate the uniform player/uniform ball distribution
    for (int g = 0; g < grid; g++)
      ball_dist[0][g] = 0;
    srand48(getpid());
    for (int test = 0; test < 10000000; test++) {
      double px = (drand48() - 0.5)*PITCH_LENGTH;
      double py = (drand48() - 0.5)*PITCH_WIDTH;
      double bx = (drand48() - 0.5)*PITCH_LENGTH;
      double by = (drand48() - 0.5)*PITCH_WIDTH;

      double dist2 = sqr(px - bx) + sqr(py - by);
      if (dist2 < sqr(cutoff_dist)) {
	double dist = sqrt(dist2);
	int g = int(floor(dist/cutoff_dist*grid));
	if (g >= grid)
	  g = grid - 1;
	ball_dist[0][g]++;
      }
    }

    int total = 0;
    for (int g = 0; g < grid; g++)
      total += ball_dist[0][g];

    ofstream out("b_dist_uniform");

    for (int g = 0; g < grid; g++) {
      double d = cutoff_dist*double(g)/grid;
      out << d << " " << double(ball_dist[0][g])/total << endl;
    }
  }
  catch (Exception e) {
    cerr << "caught exception "<< e << endl;
    return -1;
  }
  return 0;
}
