#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <netinet/in.h>
#include <string>
#include <SServer.h>
#include <sys/stat.h>
#include <errno.h>

#include <MsgStore.h>
#include "FunsselStrings.h"

#include "dribbleAnalyze.h"

using namespace std;

// cwz-build-command: make TARGET=dribbleAnalyze

// a class writing out .fig-pictures. Good for debugging/optimizing
// Based on Miha's Positions.C code from RoboCup 98

class FigDraw
{
  ofstream    *out;
public:
  FigDraw(): out(0) {}
  ~FigDraw() { delete out; }

  // creating new files
  void open(const char *fn);

  // elementary functions
  void Circle(const Vector2d &p, double rad, int color = -1, int muster = -1);
  void Line(const Vector2d &p, const Vector2d &v, int color = -1);
  void Text(const Vector2d &p, const char *txt);
  void Text(const char *txt);
  // print two numerical values (vector in some coordinates, ...)
  void VecValue(const char *vecname, double x, double y);
  
private:
  
  static const int FigDrawFactor = 2400;
  static const int OffsetX        = 1*FigDrawFactor;
  static const int OffsetY        = 1*FigDrawFactor;
  
  double line;
} drawing;

void FigDraw::open(const char *fn)
{
  delete out;
  out = new ofstream(fn);
  *out << "#FIG 3.2" << endl
       << "Landscape" << endl
       << "Center" << endl
       << "Metric" << endl
       << "A4" << endl
       << "100.00" << endl
       << "Single" << endl
       << -2 << endl
       << "1200 2" << endl;
  line = 0.0;
}

void FigDraw::Circle(const Vector2d &p, double r, int color, int ms)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());
  int mr = int(FigDrawFactor*r);
  
  *out << "1 3 0 1 " << color << " 0 0 0 "
       << ms << " 0.000 1 0.0000 "
       << mx << " " << my << " " << mr << " " << mr << " " << mx << " " << my
       << " " << mx+150 << " " << my+150 << endl;
}

void FigDraw::Line(const Vector2d &p, const Vector2d &v, int color)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());
  int mx2 = OffsetX + int(FigDrawFactor*(p.getX()+v.getX())); 
  int my2 = OffsetY - int(FigDrawFactor*(p.getY()+v.getY()));
  
  *out << "2 1 0 1 "  << color << " 0 0 -1 0.000 0 0 -1 1 0 2" << endl
       << "0 0 1.00 60.00 120.00" << endl
       << mx << " " << my << " " << mx2 << " " << my2 << " " << endl;
}

void FigDraw::Text(const Vector2d &p, const char *txt)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());

  *out << "4 0 -1 0 0 0 18 0.0000 4 180 1425 " << mx << " "
       << my << " " << txt << "\\001\n"
       << endl;
}

void FigDraw::Text(const char *txt)
{
  Text(Vector2d(1, -(line += 0.15)), txt);
}

inline void FigDraw::VecValue(const char *text,
			      double a, double b)
{
  char buf[256];
  sprintf(buf, "%s (%f,%f)", text, a, b);
  drawing.Text(buf);
}

//////////////////////////////////////////////////////////////////
// code
//////////////////////////////////////////////////////////////////

const double kick_dist = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;

int lastDribblePlayer;
Vector2d lastPlayerPos;
Vector2d lastBallPos;

Vector2d startpos;

int playerNum(SoccerActor actor)
{
  if ((int)actor >= 1 && (int)actor <= 11)
    return (int)actor - 1;
  if ((int)actor >= 13 && (int)actor <= 23)
    return (int)actor - 2;
  return -1;
}

int main(int argc, const char **argv)
{
  FigDraw log;
  ofstream glog("dribble.log");

  try {
    for (int arg = 1; arg < argc; arg++) {
      cerr << "analyze " << argv[arg] << endl;
      glog << "#######################################" << endl;
      glog << "# " << argv[arg] << endl;

      cerr << 100*arg/argc << "% done" << endl;
      MsgStore store(argv[arg]);

      lastDribblePlayer = -1;
      for (int time = 0; time <= 6000; time++) {
	if (!store.fetchTime(time))
	  break;

	if ((time % 100) == 0)
	  cerr << "time " << time << "\r";

	SoccerMessage msg;
	SoccerMessage info_msg;
	SServer::showinfo_t2 *info = 0;
	int dribblePlayer = -1;
	char dribbleAction = 'n';
	while ((msg = store.getNextMessage()).isValid()) {
	  SServer::showinfo_t2 *t_info = getShowInfo(msg);
	  if (t_info) {
	    info_msg = msg;
	    info = t_info;
	  }
	  if (!strncmp(msg.message.c_str(), FILTER_STRING, FILTER_STRING_LEN)) {
	    // try to find out who dribbles
	    istringstream m(msg);
	    string strdta;
	    int time;

	    m.seekg(FILTER_STRING_LEN);
	    m >> time;
	    m >> strdta;
	    if (strdta == "(log") {
	      m >> strdta;
	      if (strdta == "dribbleanalyze:") {
		if (dribblePlayer == -1) {
		  dribblePlayer = playerNum(msg.sender);
		  m >> strdta;
		  if (!strncmp(strdta.c_str(), "kick", 4))
		    dribbleAction = 'k';
		  else if (!strncmp(strdta.c_str(), "turn", 4))
		    dribbleAction = 't';
		  else if (!strncmp(strdta.c_str(), "dash", 4))
		    dribbleAction = 'd';
		}
		else {
		  glog << endl << setw(4) << setfill('0') << time << "WARNING: multiple dribblers, results unreliable" << endl;
		}
	      }
	    }
	  }
	}

	if (info) {
	  if (lastDribblePlayer >= 0) {
	    log.Circle(getPlayerPos(info, lastDribblePlayer) - startpos, PLAYER_SIZE, 0, 0);
	    log.Circle(getBallPos(info) - startpos, BALL_SIZE, 1, 0);
	    ostringstream tmp;
	    tmp << "dist=" << norm(getBallPos(info) - getPlayerPos(info, lastDribblePlayer));
	    log.Text(getPlayerPos(info, lastDribblePlayer) - startpos + Vector2d(0, 0.15), tmp.str().c_str());
	    log.Line(lastPlayerPos - startpos,
		     getPlayerPos(info, lastDribblePlayer) - lastPlayerPos, 0);
	    log.Line(lastBallPos - startpos, getBallPos(info) - lastBallPos, 0);
	    lastPlayerPos = getPlayerPos(info, dribblePlayer);
	    lastBallPos = getBallPos(info);
	  }

	  if (lastDribblePlayer != dribblePlayer) {
	    if (lastDribblePlayer > 0) {
	      glog << " dist " << norm(getPlayerPos(info, lastDribblePlayer) - startpos);
	      glog << endl;
	    }

	    if (dribblePlayer != -1) {
	      ostringstream tmp;
	      tmp << "dribble/dribble_" << setw(4) << setfill('0') << info_msg.time
		  << "_" << dribblePlayer;
	      if (norm(getBallPos(info) - getPlayerPos(info, dribblePlayer)) > kick_dist)
		tmp << "_noball";
	      tmp << ".log";
	      log.open(tmp.str().c_str());

	      // write initial position
	      startpos = getPlayerPos(info, dribblePlayer);
	      log.Circle(Vector2d(0, 0), PLAYER_SIZE, 0, 0);
	      log.Circle(getBallPos(info) - startpos, BALL_SIZE, 1, 0);
	      tmp.seekp(0);
	      tmp << "Start player " << dribblePlayer << (char)0;
	      log.Text(Vector2d(0, 0), tmp.str().c_str());
	      log.Text(getBallPos(info) - startpos, "Start");
	      tmp.seekp(0);
	      tmp << "dist=" << norm(getBallPos(info) - getPlayerPos(info, dribblePlayer)) << (char)0;
	      log.Text(Vector2d(0, 0.15), tmp.str().c_str());
	      lastPlayerPos = getPlayerPos(info, dribblePlayer);
	      lastBallPos = getBallPos(info);

	      glog << setw(4) << setfill('0') << info_msg.time << " player" << dribblePlayer << " ";
	    }
	    lastDribblePlayer = dribblePlayer;
	  }
	  
	  if (dribblePlayer != -1)
	    glog << dribbleAction;
	}
      }
      cerr << "done" << endl;
    }
  }
  catch (Exception e) {
    cerr << "caught exception "<< e << endl;
    return -1;
  }
  return 0;
}
