#include <iostream>
#include <Grid.h>
#include <fstream>
#include <strstream>
#include "showGoalTrace.h"
#include <MsgStore.h>
#include <netinet/in.h>
#include <string>
#include <SServer.h>
#include <sys/stat.h>
#include <errno.h>

// cwz-build-command: make analyze

const int STEPS = 20;

// statics
int attached;
string team[2];
int    team_cnt[2] = { 0, 0 };
// counter for data files
int file = 0;

bool tryAttachTeams(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return false;

  // check wether team names set
  if (show2->team[0].name[0])
    team[0] = show2->team[0].name;
  if (show2->team[1].name[0])
    team[1] = show2->team[1].name;
  return (show2->team[0].name[0] &&
	  show2->team[1].name[0]);
}

int shotGoal(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return -1;

  if (show2->pmode == SServer::PM_AfterGoal_Left)
    return 0;
  else if (show2->pmode == SServer::PM_AfterGoal_Right)
    return 1;
  return -1;
}

void analyzeGoal(int teamID, MsgStore &store)
{
  int goal_time = store.getCurrentFetchTime();
  cerr << "analyzing goal at " << goal_time
       << " for team " << team[teamID] << endl;

  strstream fname;
  fname << team[teamID] << "_bpos" << team_cnt[teamID] << (char) 0;
  // check if file exists
  struct stat buf;
  while ((stat(fname.str(), &buf) == 0) ||
	 (errno != ENOENT)) {
    team_cnt[teamID]++;
    fname.seekp(0, ios::beg);
    fname << team[teamID] << "_bpos" << team_cnt[teamID] << (char) 0;
  }

  ofstream ballpos(fname.str(), ios::app);
  fname.seekp(0, ios::beg);
  fname << team[teamID] << "_bvel" << team_cnt[teamID] << (char) 0;
  ofstream ballvel(fname.str(), ios::app);
  fname.seekp(0, ios::beg);
  fname << team[teamID] << "_kpos" << team_cnt[teamID] << (char) 0;
  ofstream kickpos(fname.str(), ios::app);
  team_cnt[teamID]++;

  int trackplayer = -1;
  // scan until ball in other half
  for (int time = goal_time - 1; time > 0; time--) {
    store.fetchTime(time);
    SoccerMessage msg = store.getNextMessage();
    SServer::showinfo_t2 *show2;
    while (!(show2 = getShowInfo(msg))) {
      msg = store.getNextMessage();
      if (!msg.isValid())
	break;
    }
    // if there is now showinfo for that cycle, ignore
    if (!show2)
      continue;

    Vector2d bpos = getBallPos(show2);
    Vector2d bvel = getBallVel(show2);

    ballpos << bpos.getX() << " " << bpos.getY() << endl;

    ballvel << bpos.getX() << " " << bpos.getY() << endl
	    << bpos.getX() + bvel.getX() << " "
	    << bpos.getY() + bvel.getY() << endl << endl;

    for (int i = 0; i < 11; i++)
      if (playerKicked(show2, 11*teamID + i))
	trackplayer = 11*teamID + i;

    if (trackplayer != -1) {
      Vector2d ppos = getPlayerPos(show2, trackplayer);
      kickpos << ppos.getX() << " "
	      << ppos.getY() << endl;
    }
    
    if ((bpos.getX() < 10 && teamID == 0) ||
	(bpos.getX() > -10 && teamID == 1))
      break;
  }
}

int main(int argc, const char **argv)
{

  if (argc < 2) {
    cerr << "wrong #params." << endl;
    return -1;
  }

  for (int arg = 1; arg < argc; arg++) {
    MsgStore store(argv[arg]);
    team[0] = team[1] = "NONE";
    attached = 0;

    for (int time = 0;; time++) {
      if (!store.fetchTime(time))
	break;

      if ((time % 100) == 0)
	cerr << "time " << time << "\r";

      SoccerMessage msg;
      while ((msg = store.getNextMessage()).isValid()) {
	if (!attached)
	  if (tryAttachTeams(msg))
	    attached = 1;
	if (attached) {
	  switch (shotGoal(msg)) {
	  case  0:
	    analyzeGoal(0, store);
	  break;
	  case  1:
	    analyzeGoal(1, store);
	  break;
	  default: break;
	  }
	}
      }
    }
    cerr << "done" << endl;
  }
  return 0;
}
