#include <iostream>
#include <iomanip>
#include <Grid.h>
#include <fstream>
#include <strstream>
#include "showScores.h"
#include <MsgStore.h>
#include <netinet/in.h>
#include <string>
#include <SServer.h>
#include <sys/stat.h>
#include <errno.h>

// cwz-build-command: gmake showScores

// team identification
string team[2];
int    team_id[2];

// scores
const int n_teams = 10;
const char *teams[n_teams] = {
  "11MonkeysZ00",
  "ATH2000neu",
  "BS2001",
  "Cyberoos01",
  "FCPortugal00",
  "MRB01",
  "Robolog01",
  "TutGroove01",
  "VirtualWerder01",
  "YowAI00" 
};

int score[n_teams][n_teams] = { { 0 } };
int events[n_teams][n_teams] = { { 0 } };

void tryAttachTeams(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return;

  team_id[0] = team_id[1] = -1;
  if (show2->team[0].name[0]) {
    team[0] = show2->team[0].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[0] == teams[t] ||
	  team[0] == string(teams[t]) + "-2")
	team_id[0] = t;
    }
  }
  if (show2->team[1].name[0]) {
    team[1] = show2->team[1].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[1] == teams[t] ||
	  team[1] == string(teams[t]) + "-2")
	team_id[1] = t;
    }
  }
}

int shotGoal(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return -1;

  if (show2->pmode == SServer::PM_AfterGoal_Left)
    return 0;
  else if (show2->pmode == SServer::PM_AfterGoal_Right)
    return 1;
  return -1;
}

int main(int argc, const char **argv)
{
  try {
    if (argc < 3 || strcmp(argv[2], "-logs")) {
      cerr << "wrong #params." << endl;
      return -1;
    }

    cout.precision(2);
    
    for (int arg = 3; arg < argc; arg++) {
      MsgStore store(argv[arg]);

      // search last time step and get showinfo
      store.fetchTime(6000);
      SoccerMessage msg;
      while ((msg = store.getNextMessage()).isValid()) {
	tryAttachTeams(msg);
	if (team_id[0] != -1 && team_id[1] != -1) {
	  // attached, so we have got a showinfo and
	  // hence the team names and score
	  events[team_id[0]][team_id[1]]++;
	  events[team_id[1]][team_id[0]]++;
	  
	  SServer::showinfo_t2 *show2 = getShowInfo(msg);
	  if (!show2)
		continue;
 
	  score[team_id[0]][team_id[1]] += ntohs(show2->team[0].score);
	  score[team_id[1]][team_id[0]] += ntohs(show2->team[1].score);
	  break;
	}
      }
      
      // write intermediate result
      cout << setw(16) << "team";
      for (int a = 0; a < n_teams; a++)
	cout << " " << teams[a];
      cout << endl;
      
      for (int a = 0; a < n_teams; a++) {
	cout << setw(16) << teams[a];
	for (int b = 0; b < n_teams; b++) {
	  if (events[a][b])
	    cout << setw(strlen(teams[b]) + 1) << double(score[a][b])/events[a][b];
	  else
	    cout << setw(strlen(teams[b]) + 1) << "*";
	}
	cout << endl;
      }
    }
    
    // write final result
    ofstream out(argv[1]);
    out.precision(2);
    
    out << setw(16) << "team";
    for (int a = 0; a < n_teams; a++)
      out << " " << teams[a];
    out << endl;
    
    for (int a = 0; a < n_teams; a++) {
      out << setw(16) << teams[a];
      for (int b = 0; b < n_teams; b++) {
	if (events[a][b])
	  out << setw(strlen(teams[b]) + 1) << double(score[a][b])/events[a][b];
	else
	  out << setw(strlen(teams[b]) + 1) << "*";
      }
      out << endl;
    }

    out << endl << "events measured" << endl << setw(16) << "team";
    for (int a = 0; a < n_teams; a++)
      out << " " << teams[a];
    out << endl;
    
    for (int a = 0; a < n_teams; a++) {
      out << setw(16) << teams[a];
      for (int b = 0; b < n_teams; b++)
	out << setw(strlen(teams[b]) + 1) << events[a][b];
      out << endl;
    }

    return 0;
  }
  catch (Exception e) {
    cerr << "got exception " << e << endl;
    return -1;
  }
}
