//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Trace.h"
#include "Comm.h"
#include "Config.h"
#include "minmax.h"
#include "mcommand.h"
#include <string.h>
#include <iostream>

using namespace std;

Trace * theTrace = 0;

void theTrace_init( bool FUNSSELMode, const char * configFile ) {
  if (!theTrace)
    theTrace = new Trace(FUNSSELMode, configFile);
}

void theTrace_stop( void ) {
  delete theTrace;
  theTrace = 0;
}

Trace::ModuleStrings_t Trace::moduleStrings[NUM_MODULES] = {
  {MC,     "MC_players",     "MC_level",     "MC_min_level",     "MC_draw",     "MasterControl"},
  {WM,     "WM_players",     "WM_level",     "WM_min_level",     "WM_draw",     "Worldmodel"},
  {TACT,   "TACT_players",   "TACT_level",   "TACT_min_level",   "TACT_draw",   "Tactical Module"},
  {HAND,   "HAND_players",   "HAND_level",   "HAND_min_level",   "HAND_draw",   "Ballhandling Module"},
  {PASS,   "PASS_players",   "PASS_level",   "PASS_min_level",   "PASS_draw",   "Pass Module"},
  {GOAL,   "GOAL_players",   "GOAL_level",   "GOAL_min_level",   "GOAL_draw",   "Goalshot Module"},
  {POS,    "POS_players",    "POS_level",    "POS_min_level",    "POS_draw",    "Position Module"},
  {STDSIT, "STDSIT_players", "STDSIT_level", "STDSIT_min_level", "STDSIT_draw", "Standard Situation Module"},
  {SKILLS, "SKILLS_players", "SKILLS_level", "SKILLS_min_level", "SKILLS_draw", "Advanced Skills"},
};


Trace::Trace( bool FMode, const char * configFile )
  : writeToFile(false),
    traceFile(),
    of(),
    FUNSSELMode(FMode),
    isActive(false),
    configFile(configFile),
    player(-1)
{
  const char * value;

  if (!Config::readFile(configFile)) {
    cerr << "Trace: configfile " << configFile<< " not found" << endl;
    exit(-1);
  }

  if (!(value = Config::getValue("logToFile"))) {
    cerr << "Trace: missing 'logToFile' in " << configFile << endl;
    exit(-1);
  };
  
  writeToFile = boolStringValue(value);
  if (writeToFile) {
    if (!(value = Config::getValue("filePrefix"))) {
      cerr << "Trace: missing 'filePrefix in' " << configFile << endl;
      exit(-1);
    };
    traceFile = value;
  }

  for (int i = 0; i < NUM_MODULES; ++i) {
    if (!(value = Config::getValue(moduleStrings[i].playersStr))) {
      cerr << "Trace: missing '" << moduleStrings[i].playersStr 
	   << "' in " << configFile << endl;
      exit(-1);
    };
    setModulePlayerInfo(value, moduleStrings[i].module);

    if (!(value = Config::getValue(moduleStrings[i].levelStr))) {
      cerr << "Trace: missing '" << moduleStrings[i].levelStr 
	   << "' in " << configFile << endl;
      exit(-1);
    };
    setModuleLevelInfo(value, modules[moduleStrings[i].module].level);
  
    if (!(value = Config::getValue(moduleStrings[i].min_levelStr))) {
      cerr << "Trace: missing '" << moduleStrings[i].min_levelStr 
	   << "' in " << configFile << endl;
      exit(-1);
    };
    setModuleLevelInfo(value, modules[moduleStrings[i].module].min_level);
  
    if (!(value = Config::getValue(moduleStrings[i].drawStr))) {
      cerr << "Trace: missing '" << moduleStrings[i].drawStr 
	   << "' in " << configFile << endl;
      exit(-1);
    };
    modules[moduleStrings[i].module].draw = boolStringValue(value);
  }
}

void
Trace::setPlayer( int p, int s ) {
  if (!((0 <= p) && (p <= 12))) {
    cerr << "Trace: invalid player number " << p << endl;
    exit(-1);
  };

  if (writeToFile) {
    char num[10];

    sprintf(num, "-%c-%02d.log", (s == 0) ? 'l' : 'r', p);
    traceFile += num;
    of.open(traceFile.c_str());
    cerr << "Trace("<< p << "): writing to file " << traceFile << endl;
  }

  if (p == 0) // coach is number 12 here
    p = 12;
  player = (p > 1) ? (2 << (p-2)) : 1;

  for (int i = 0; i < NUM_MODULES; ++i)
    if (isActive = (player & modules[i].players))
      break;
}

bool
Trace::boolStringValue( const char * str ) {
  return !strcmp(str, "true") || !strcmp(str, "1") ||
    !strcmp(str, "on") || !strcmp(str, "yes") || 
    !strcmp(str, "y");
}

void
Trace::setModulePlayerInfo( const char * info, Module module ) {
  if (!strcmp(info, "all"))
    modules[module].players = 0xFFF; // 2^12 - 1
  else if (!strcmp(info, "none") || !strcmp(info, "0"))
    modules[module].players = 0;
  else {
    int result;
    modules[module].players = 0;
    while ((result = getPlayers(info)) != 0) {
      if (!((0 < result) && (result <= 12))) {
	cerr << "Trace: invalid format for " << moduleStrings[module].playersStr 
	     << " in " << configFile << endl;
	exit(-1);
      };
      modules[module].players |= (result > 1) ? (2 << (result-2)) : 1;
    }
  }
}

int
Trace::getPlayers( const char * &line ) {
  int result = -1, read = 0;

  if (strlen(line) <= 0)
    return 0;

  if (sscanf(line, ";%d%n", &result, &read) == 1) {
    line += read;
    return result != 0 ? result : 12;
  }

  if (sscanf(line, "%d%n", &result, &read) == 1) {
    line += read;
    return result != 0 ? result : 12;
  }
  
  return -1;
}

void
Trace::setModuleLevelInfo( const char * info, unsigned int & level ) {
  int offset;

  if (!strcmp(info, "DECISION"))
    level = DECISION;

  else if (!strcmp(info, "CONDITION"))
    level = CONDITION;

  else if (!strcmp(info, "VALUE"))
    level = VALUE;

  else if (sscanf(info, "DECISION%d", &offset) == 1)
    level = max(0, DECISION + offset);

  else if (sscanf(info, "CONDITION%d", &offset) == 1)
    level = max(0, CONDITION + offset);

  else if (sscanf(info, "VALUE%d", &offset) == 1)
    level = max(0, VALUE + offset);

  else if (sscanf(info, "%d", &offset) == 1)
    level = max(0, offset);
  
  else
    level = 0;
}

void
Trace::write( Module mod, unsigned int level, 
	      const char * file, const char * function, const char * line,
	      ostringstream & str )
{
  if (player & modules[mod].players &&
      modules[mod].min_level <= level && level <= modules[mod].level &&
      FUNSSELMode)
    {
      MELog log(mod, level, file, function, line, str.str());
      if (writeToFile) {
	of << log << endl;
      } else {
	Comm::sendMCmd(log);
      }
    }
}

void
Trace::write( const Worldmodel & world ) {
  if (isActive && FUNSSELMode) {
    MSnapshot log(world);
    if (writeToFile) {
      of << log << endl;
    } else {
      Comm::sendMCmd(log);
    }
  }
}

void
Trace::write( Module mod, unsigned int level, const MCommand & log ) {
  if (player & modules[mod].players && modules[mod].draw && 
      modules[mod].min_level <= level && level <= modules[mod].level &&
      FUNSSELMode)
    {
      if (writeToFile) {
	of << log << endl;
      } else {
	Comm::sendMCmd(log);
      }
    }
}

void
Trace::write( Module mod, const MCommand & log ) {
  if (player & modules[mod].players && FUNSSELMode)
    {
      if (writeToFile) {
	of << log << endl;
      } else {
	Comm::sendMCmd(log);
      }
    }
}
