//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "coordinates.h"
#include "mcommand.h"
#include "worldmodel.h"

#include <iostream>
#include <string>
#include <sstream>
#include "stdio.h"

#include "FunsselStrings.h"

using namespace std;
// class MCommand

ostream& operator<< (ostream& os, const MCommand &mc)
{
  return os << string(mc);
}

MCommand::MCommand( int duration, const string &encoded)
  : duration(duration) 
{
  ostringstream ost;
  ost << encoded;
  const char *str= ost.str().c_str();
  this->encoded= str;
}

// class MMarker
MMarker::MMarker( const Vector2d &pos, const char *label, int duration)
  : MCommand( duration )
{
  char buf[201];
  const char *tmp= label;
  if (label == 0)
    tmp= "";

  sscanf(tmp, "%200[^\"]", buf);

  ostringstream ost;
  ost << FMC_MARK " " << duration << " " 
      << pos << '"' << buf << '"' << " )";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MLog
MLog::MLog( const char *text, int duration )
  : MCommand(duration, string(FMC_LOG " "+string(text)+string(")") ))
{}

MLog::MLog( const std::string text, int duration )
  : MCommand(duration, FMC_LOG " "+text+")")
{}

// class MELog
MELog::MELog( int module, int level,
	      const char * file, const char * function, const char * line, 	      
	      const char * text, int duration)
  : MCommand(duration)
{
  char buf[7];
  sprintf(buf, "%2d %3d", module, level);
  encoded = string(FMC_ELOG +string(buf)+
		   "\""+string(file)+"\""+
		   "\""+string(function)+"\""
		   +string(line)+" "+string(text) );
}

MELog::MELog( int module, int level,
	      const char * file, const char * function, const char * line, 	      
	      string text, int duration)
  : MCommand(duration)
{
  char buf[7];
  sprintf(buf, "%2d %3d", module, level);
  encoded = string(FMC_ELOG +string(buf)+
		   "\""+string(file)+"\""+
		   "\""+string(function)+"\""
		   +string(line)+" "+string(text) );
}

// class MSnapshot
MSnapshot::MSnapshot( const Worldmodel & world )
{
  ostringstream ost;
  ost << FMC_SNAP " ";
  world.writeToStream( ost );
  ost << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}  

// class MDCircle
MDCircle::MDCircle(const Vector2d &pos, double rad, bool hasGrid, 
		   MStyle style,  int duration)
  : MCommand(duration)
{
  ostringstream ost;
  ost << FMC_CIRCLE " " << duration << " " << pos << " " 
      << rad << " " << hasGrid << " " << style << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// MDPosition
MDPosition::MDPosition(const Vector2d &pos, int duration)
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_POS " " << duration << " " << pos << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MDLine
MDLine::MDLine(const Vector2d &start, const Vector2d &end, MStyle style, int duration)
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_LINE " " << duration << " " << start 
      << " " << end << " " << style << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MDVLine
MDVLine::MDVLine(double x, MStyle style, int duration)
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_VLINE " " << duration << " " << x << " " << style  << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MDBox
MDBox::MDBox(const Vector2d &start, const Vector2d &end, MStyle style, int duration)
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_BOX " " << duration << " " << start 
      << " " << end << " " << style << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MDCorridor 
MDCorridor::MDCorridor( const Vector2d &source, const Vector2d &target,
			double diam, const Angle &alpha, MStyle style,
			int duration)
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_CORR " " << duration << " " << source << " " << target 
      << " " << diam << " " << alpha << " " << style  << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MDLabel
MDLabel::MDLabel( const Vector2d &pos, const string &text, int duration )
  : MCommand( duration )
{
  ostringstream ost;
  ost << FMC_LABEL " " << duration << " " << pos << " " << text << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MStop

MStop::MStop() 
  : MCommand( 1, string(FMC_STOP ")") )
{}

// class MDropBall
MDropBall::MDropBall( const Vector2d &pos ) 
{
  ostringstream ost;
  ost << FMC_DROPBALL " " << duration << " " << pos << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MFreekickUs
MFreekickUs::MFreekickUs( const Vector2d &pos )
{
  ostringstream ost;
  ost << FMC_FREEKUS " " << duration << " " << pos << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MFreekickThem
MFreekickThem::MFreekickThem( const Vector2d &pos )
{
  ostringstream ost;
  ost << FMC_FREEKTHEM " " << duration << " " << pos << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}

// class MSetPlayer
MSetPlayer::MSetPlayer( bool ourTeam, int playerNo, const Vector2d &pos, 
			const Angle &ang)
{
  ostringstream ost;
  ost << FMC_SETPLYR " " << duration << " " << playerNo << " ";
  if (ourTeam)
    ost << 1;
  else
    ost << 0;
  ost << " " << pos << " " << ang << ")";
  const char *str= ost.str().c_str();
  encoded = str;
}
