//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef _ONLINECOACH_H_
#define _ONLINECOACH_H_

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>


#include "constants.h"
#include "Comm.h"
#include "worldmodel.h"
#include "audiomgr.h"

// filename of worldmodel log files
#define WORLD_LOG_FILENAME "world_"
#define WORLD_LOG_EXTENSION ".log"

class Onlinecoach {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in an onlinecoach client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.

  Onlinecoach (const char*  team_name,       // name of own team
	       const char*  host             = DEFAULT_HOST,
	       int          port             = OLCOACH_PORT,   
	       const char*  configfile       = "agent.conf",
	       const char*  traceCfg         = "trace.conf",
	       bool         FUNSSELmode      = false);

  // Starts execution of the Onlinecoach client. 
  // After calling this method, the agent will repeatedly call act()
  // every time it needs to send an action while the server is running.
  //
  // You cannot override this method.
  void coach();

  
  // Called once every time the online coach can say something to the team,
  // see also analyze() below.
  //
  // Override this method virtual in your derived classes to implement 
  // onlinecoach behaviour.
  virtual void act() = 0;

  // Called once every timestep during the game. If the game timer is 
  // stopped, it is repeatedly called once every SIMULATOR_STEP ms.
  // It is always called _before_ act() for the relevant timestep.
  //
  // Override this method virtual in your derived classes to implement 
  // onlinecoach behaviour.
  virtual void analyze() = 0;

  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Onlinecoach();  


  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();}
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}

  // returns the timestep the server is in, our current time.
  // (There is only ONE current time. This is it.)
  long         getTime() {return comm->getServerTime();} 

  // sends an audiomessage to other clients immediately
  void         say(AudioMsg &msg) {comm->say(msg);}

  // sends "(done)" to the server needed if synch_mode
  void         sendDone( void ) {comm->sendDone();};    

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  // send a change player type msg to sserver
  void sendChangePlayer(int player, int type) 
  { comm->sendChangePlayer(player, type); };

  Worldmodel *world;

private:

  Comm *comm;

};


#endif




