//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __GOALSHOTEVAL_H__
#define __GOALSHOTEVAL_H__

#include "grade.h"
#include "mcommand.h"
#include "worldmodel.h"
#include "Config.h"

#include <sstream>


//////////////////////////////////////////////////////////////////////
//
// class GoalShotEval
//
//////////////////////////////////////////////////////////////////////

class GoalShotEval
{
public:
  GoalShotEval();

  ~GoalShotEval() {}
  
  // do the evaluation for a player of our team
  // for definition of Grade: see grade.h
  Grade ourEval( const Worldmodel * world, const Vector2d &playerpos ) const;

  // do the evaluation for a player of the other team
  // (does not use world->ball, only world->ourTeam[] and playerpos)
  Grade otherEval( const Worldmodel * world, const Vector2d &playerpos ) const;

private:

  // evaluate one of the teams (depending on ourTeamGoalShot)
  // bestTarget holds the target on the goal line player should aim at
  Grade teamEval( const Worldmodel * world, const Vector2d &playerpos, Vector2d &bestTarget, bool ourTeamGoalShot ) const;

  // edges of opponent goal
  Vector2d leftEdgeOfGoal;
  Vector2d rightEdgeOfGoal;

  // edges of our goal
  Vector2d leftEdgeOfOurGoal;
  Vector2d rightEdgeOfOurGoal;

  // distances to the edges of the goal where the player should kick
  // to - depends on the distance between the player and the goal
  double minDistToCorner;
  double maxDistToCorner;

  // maximum time to do a kick
  int maxKickDuration;

  // maximum score for a target to be graded with POOR
  int score_POOR;

  // maximum score for a target to be graded with AVERAGE
  int score_AVERAGE;

  // maximum score for a target to be graded with GOOD
  int score_GOOD;

  // maximum score for a target to be graded with PERFECT
  int score_PERFECT;

  // maximum number of simulation steps in goalShotEval()
  int maxSimSteps;

  // maximum dist for a goalshot (is calculated from maxSimSteps)
  double maxDist;

  // older objects will not be considered
  int maxAge;   

  // calculates a correction dist to the corner of the goal where the
  // player finally should kick to
  Vector2d correctCorner( double dist ) const;

  friend class GoalShotModule;   // this module will use some of the private methods
 
}; // class GoalShotEval

#endif
