//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __PASSEVAL_H__   
#define __PASSEVAL_H__

#include "grade.h"
#include "worldmodel.h"
#include "tacticalmodule.h"
#include "goalshoteval.h"
#include "AdvSkills.h"

#define FUNSSELMACROS
#include "drawMacros.h"
#include "mcommand.h"
#include <sstream>
#include <iostream>


///////////////////////////////////////////////////////////////////
// 
// traces
//
///////////////////////////////////////////////////////////////////

#ifdef MRBDEBUG

// comment out the following line to disable trace
//#define USEPASSETRACE

#endif

#ifdef USEPASSETRACE
#define PASSETRACE(cmd) {ostringstream ost;ost << cmd << char(0); std::string str = ost.str(); Comm::sendMCmd(MLog(str));}
#define PASSECMD(cmd) cmd
#warning PassModule debug trace PASSETRACE is activated!
#else
#define PASSETRACE(cmd)
#define PASSECMD(cmd)
#endif


//////////////////////////////////////////////////////////////////////
//
// class PropOf
//
//////////////////////////////////////////////////////////////////////

class PropOf
{
public:
  int      number;      // number of partner
  Vector2d pos;         // position to pass to; not necessarily partner's position!
  int      xIndex;      // indicates grid position of pass
  int      yIndex;
  double   dist;        // distance between player and target position
  double   vel;         // velocity that ball should reach at target position
  bool     offside;     // partner is not in an offside positon
  int      score;       // score of this pass
  int      tactscore;   // tactical score of this pass (is it good to play this pass?)
  int      corrscore;   // corridor score of this pass (is it possible to play this pass?)
  int      agescore;    // is the information on the world to old ?
  int      numSteps;    // number of steps until partner reaches interceptpos
  int      duration;    // duration of the pass
  Grade    grade;       // grade of this pass

  // shout the PropOf
  friend std::ostream& operator << ( class std::ostream &os, const PropOf prop )
  {
    return os << " pass to  " << prop.number
	      << " at pos " << prop.pos
	      << " dist:" << prop.dist
	      << " vel:" << prop.vel
	      << " offside:" << prop.offside
	      << " score:" << prop.score
	      << " tscore:" << prop.tactscore
	      << " cscore:" << prop.corrscore
	      << " ascore:" << prop.agescore
	      << " steps:" << prop.numSteps
	      << " dur:" << prop.duration
	      << " grade:" << prop.grade
	      << " x:" << prop.xIndex
	      << " y:" << prop.yIndex;
  }
};


//////////////////////////////////////////////////////////////////////
//
// class PassEval
//
//////////////////////////////////////////////////////////////////////

class PassEval
{
public:
  PassEval( const GoalShotEval* _goalshoteval);
  ~PassEval() {}

  // this methods needs a worldmodel, a playernumber and a partnernumber;
  // it grades the passes from player to partner on the basis of the worldmodel,
  // considering also rangepasses;
  // the position where the pass would be played to is reported in posOfPass
  Grade ourEval( const Worldmodel* world,  const TacticalModule * tact,
		 int plnumber, int panumber, Vector2d &posOfPass ) const
  { PropOf prop = ourEvalProp( world, tact, plnumber, panumber ); posOfPass = prop.pos; return prop.grade; };

  // just for calling the above method without posOfPass
  Grade ourEval( const Worldmodel* world, const TacticalModule * tact, int plnumber, int panumber ) const
  { PropOf prop = ourEvalProp( world, tact, plnumber, panumber ); return prop.grade; };

  // does the same as above but reports PropOf of best pass
  PropOf ourEvalProp( const Worldmodel* world, const TacticalModule * tact, int plnumber, int panumber ) const;

  // this methods does the same, but for players of the other team;
  // it takes no features of the other team into account; also does not consider rangepasses
  Grade otherEval( const Worldmodel* world, const TacticalModule * tact, int plnumber, int panumber ) const;

  // maximum scores for a pass to be graded with this grade
  int SCORE_POOR;
  int SCORE_AVERAGE;
  int SCORE_GOOD;
  int SCORE_PERFECT;

  // calculate Grade from a score
  Grade scoreToGrade( int score ) const
  { return ( score <= SCORE_PERFECT ? Grade::PERFECT :
	     score <= SCORE_GOOD ? Grade::GOOD :
	     score <= SCORE_AVERAGE ? Grade::AVERAGE :
	     score <= SCORE_POOR ? Grade::POOR :
	     Grade::BAD ); };

  // calculate (maximum) score from a Grade
  int   gradeToScore( const Grade& grade ) const
  { return ( grade.is_PERFECT() ? SCORE_PERFECT :
	     grade.is_GOOD() ? SCORE_GOOD :
	     grade.is_AVERAGE() ? SCORE_AVERAGE :
	     grade.is_POOR() ? SCORE_POOR :
	     SCORE_POOR + 1 ); };

private:
  // this module is used by the passmodule of a player, too!
  friend class PassModule;

  const GoalShotEval* goalshoteval;

  // to decide when an player is on the same level with another player
  double SAMELEVEL_LENGTH;

  int    MAXKICK_DURATION;      // maximal time for a kick

  double MAXPASS_LENGTH;        // maximal length of a pass

  double RPASS_LATTICELENGTH;   // distance between rangepass grid points

  // length of a pass that is too short to play it if the sit. is not dangerous
  double SHORTPASS_LENGTH;

  // length of a pass that is too short to play it for the goalie
  double SHORTPASS_LENGTH_GOALIE;

  // distance to the sides of the field
  double SECURITY_DIST;

  // distance to the border of the field that is to be kept for passes
  double BORDER_DIST;

  // check how much time player has for his kick
  long timeForPass( const Worldmodel *world, int plnumber ) const;
 
  // get the PropOf of a pass from player no. "plnumber" to "prop.pos"
  // the receiver of the pass is its teammate with no. "panumber"
  // if we == true, the method works for our team, else for the opponent
#ifndef USEPASSETRACE
  bool evalPass( bool we, const Worldmodel *world, const TacticalModule * tact,
		 int plnumber, int panumber,
		 double offsideLine, long maxTimeForPass, double maxBallSpeed,
		 PropOf &prop) const;
#else
  bool evalPass( bool we, const Worldmodel *world, const TacticalModule * tact,
		 int plnumber, int panumber,
		 double offsideLine, long maxTimeForPass, double maxBallSpeed,
		 PropOf &prop, bool debug = false ) const;
#endif
};

#endif
