//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __PASSMODULE_H__   
#define __PASSMODULE_H__

#include "passeval.h"
#include "goalshoteval.h"

#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "drawMacros.h"
#include "mcommand.h"
#include <sstream>


//////////////////////////////////////////////////////////////////////
//
// class PassModule
//
//////////////////////////////////////////////////////////////////////

class PassModule
{
public:
  // constructor of module PassModule
  PassModule( Worldmodel * _world, TacticalModule * _tact, AdvSkills * _skills,
	      PassEval * _passeval, const GoalShotEval * _goalshoteval );
  ~PassModule() {}
  
  // feed information about new world and tactics into module
  bool feedbackWorld();

  // write information into String (for player communication)
  // returns best grade of evaluated passes (so that tactical module can decide)
  Grade encodeAudio( char* msg );

  // Do the evaluation for the next action
  // For definition of Grade: see grade.h
  Grade evaluate();

  // Calculate a turn head wish
  TurnHeadWish turnHeadEvaluate();
 
  // Perform the action for this module
  bool act();

private:

  Worldmodel * world;
  TacticalModule * tact;
  AdvSkills * skills;

  // recall of this module needed, if true
  bool callMeAgain;

  // we want to use our own passeval method
  PassEval* passeval;

  // and perhaps the goalshoteval method
  const GoalShotEval* goalshoteval;

  // time available for a kick
  int timeForKick;

  // information on the passtarget
  PropOf   chosenPass;

  // emergency kick
  Vector2d targetOfEmergencyPass;   // direction of emergency kick
  bool     performEmergencyPass;    // true if no pass partner was found

  // information concerning the looking after ball
  bool     lookAfterPass;       // true, if player should look after the ball
  int      lookAfterPassTime;   // time when looking for ball started
  Vector2d lookAfterPassPos;    // position player is looking at after pass

  Vector2d lookPos;      // position where the player wants to look at (pass target)
  Grade turnHeadGrade;   // grade for turning the head towards a pass target

  // information about the kick commands
  AdvSkills::ExtKickCommand kickparams;

  // calculate emergency kick away vector
  Vector2d getEmergencyPass();
};

#endif
