//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifdef MRBDEBUG
//#define POS_RES_TRACE(output) Comm::sendMCmd(output)

//#define ATTACKTRACE
//#define FREEPOSTRACE
//#define COVERTRACE
//#define SUPPTRACE

//#define CERR(output) cerr << "POSM: "<<world->time << " No. " <<me.number<<" "<<output<< endl
#endif

#ifndef POS_RES_TRACE
#define POS_RES_TRACE(output)
#else
#warning POS_RES_TRACE aktiviert!!
#define FUNSSELMACROS
#include "drawMacros.h"
#define USEPOS_RES_LOG
#endif

#ifndef USEPOS_RES_LOG
#define POS_RES_LOG(output)
#else
#warning POS_RES_LOG aktiviert !!
#define POS_RES_LOG(output) { ostrstream ost;\
	ost << "POS_RES_LOG: " << output << (char)0;\
    char * str = ost.str(); \
 Comm::sendMCmd(MLog(str)); \
 delete str; }
#endif

#ifdef SUPPTRACE
#warning: POSM debug trace SUPPTRACE is activated!
#define SUPPTRC(output) Comm::sendMCmd(output)
#else
#define SUPPTRC(output)
#endif

#ifdef COVERTRACE
#warning: POSM debug trace COVERTRACE is activated!
#define COVERTRC(output) Comm::sendMCmd(output)
#else
#define COVERTRC(output)
#endif

#ifdef ATTACKTRACE
#warning: POSM debug trace ATTACKTRACE is activated!
#define ATTACKTRC(output) Comm::sendMCmd(output)
#define ATTACKLOG(output) { ostrstream ost;\
	ost << "attack-> " << output << (char)0;\
    char * str = ost.str(); \
 Comm::sendMCmd(MLog(str)); \
 delete str; }
#else
#define ATTACKTRC(output)
#define ATTACKLOG(output)
#endif

#ifdef FREEPOSTRACE
#warning: POSM debug trace FREEPOSTRACE is activated!
#define FREEPOSTRC(output) Comm::sendMCmd(output)
#define FREEPOSLOG(output) { ostrstream ost;\
	ost << "freepos-> " << output << (char)0;\
    char * str = ost.str(); \
 Comm::sendMCmd(MLog(str)); \
 delete str; }
#else
#define FREEPOSLOG(output)
#define FREEPOSTRC(output)
#endif

#ifndef CERR
#define CERR(output)
#else
#warning POSM CERR is aktivated
#endif

#include "PositionFunctions.h"
#include "PositionEval.h"


const double COVER_MIN_DIST=-FIELD_LENGTH*1/6.;
//opponents who are at this distance or less from our goal
//have to be covered 
const double KEEPOUT_X_FRONT = 2/3.;
const double KEEPOUT_X_BACK = 2/9. ;
const double KEEPOUT_Y = 1/8. ;

const double KEEPOUT_X_FRONT_BIG = 5/6. ;
const double KEEPOUT_X_BACK_BIG = 5/18.;
const double KEEPOUT_Y_BIG = 5/32.;

const double BIGGER_KEEPOUT_AREA = 0.3;
//this part of the Playerarea should be avoided when searching free position
//if the position is far away (which means longer than BIGGER_KEEPOUT_AREA*WidthOfPlayerarea) we check a bigger area
//continuous change of size
const double FAR_AWAY = FIELD_WIDTH/4. ;
const double NOT_ACTIVE_TOO_LONG = 5;
const double NEW_GRADE_FACTOR=1;
const double OLD_GRADE_FACTOR=1;

bool PositionModule::EvalPlayerAttack()
{
  int ourTime;
  WBall ourBall;
  int distanceTime=9999;
  WBall distanceBall;
  Vector2d OppAtBallInterception;
  

  if ( tact->firstDangerousOpponent != UNKNOWN &&
       tact->firstResponsiblePlayer != UNKNOWN )
    {
      // the player to attack
      const WPlayer& opp = world->otherTeam[tact->firstDangerousOpponent];

      // where would I be able to intercept the ball?
      ourTime=tact->ourTeamReachBallTime[tact->firstResponsiblePlayer];
      ourBall=tact->ourTeamReachBallPos[tact->firstResponsiblePlayer];


      shortestTime = 0;      // time opponent will need to reach the ball
      shortestBall = world->ball;   // ball where opponent will reach it
      int age = opp.age > 4 ? 4 : opp.age;

      // worst case simulation of the opponent           
      while ( opp.distTo( shortestBall ) > opp.kickRange() + 
	      opp.vmax * ( shortestTime + age )  )
	{
	  shortestBall.simulate();
	  shortestTime++;
	}
      // this is the opponent's earliest ball intercept point

      // now try to find out when the ball will be closest to the opponents
      // current position. If I can't reach the ball there he will probably
      // just wait for the ball.
      long dummyT = shortestTime;
      WBall dummy = shortestBall;
      distanceBall = dummy;
      distanceTime = dummyT;
      double mindist = FIELD_LENGTH;
      while ( dummyT < 2 * ourTime )
	{
	  if ( opp.distTo( dummy ) < mindist )
	    {
	      mindist = opp.distTo( dummy );
	      distanceBall = dummy;
	      distanceTime = dummyT;
	    }
	  dummy.simulate();
	  dummyT++;
	};

      // the opponent is going to intercept the ball somewhere in between
      // his first intercept point and my intercept point.
      // I will now try to figure out where.

      Vector2d Opp2Ball; 
      
      if ( shortestTime <1 )
	{
	  // opponent already has the ball
	  ATTACKTRC(MLog("evalplAttack-> opponent in kickrange"));
	  simBall=shortestBall;
	  simTime=shortestTime;
	  Opp2Ball = ( simBall.pos - opp.pos );

	  if ( opp.age > 0 )
	    OppAtBallInterception = simBall.pos - 
	      0.8*opp.kickRange()/norm(Opp2Ball)*Opp2Ball;
	  else
	    OppAtBallInterception = opp.pos;
	}
      else if ( ( opp.pos - world->ball.pos ) * world->ball.v < 0 )
	{
	  // ball is going away from our opponent
	  // he is almost certainly intercepting at the very first
	  // possibility.
	  ATTACKTRC(MLog("evalplAttack-> ball is going away from opponent"));
	  simBall=shortestBall;
	  simTime=shortestTime;
	  Opp2Ball = ( simBall.pos - opp.pos );
	  OppAtBallInterception = simBall.pos - 
	    0.8*opp.kickRange()/norm(Opp2Ball)*Opp2Ball; 
	}
      /*      else if ( simOpp.number == opp.number )
	{ // is it an old opponent, i.e. what did I do before?
	  
	  ATTACKTRC(MLog("plAttack-> using old simulation"));
	  simTime=lastpredictedTime-1;
	    simBall=lastpredictedBall;
	    }*/
      else
	{
	  // no other case: just asume he will intercept the ball right 
	  // before we will be able to get it
	  ATTACKLOG("plAttack-> new opponent using shortest distance "
		    "simulation");
	    simTime=distanceTime;
	    simBall=distanceBall;
	    Opp2Ball = ( simBall.pos - opp.pos );
	    OppAtBallInterception = simBall.pos - 
	      0.8*opp.kickRange()/norm(Opp2Ball)*Opp2Ball;
	}
      
      simOpp=opp;      
      if ( opp.age > 0 )
	{ //simulate opponent on his way to the ball
	  simOpp.pos += Opp2Ball * 
	    ((double) opp.age / ((double) simTime + (double) opp.age ));
	  ATTACKTRC(CIRCLE(simOpp.pos,PLAYER_SIZE));
	}
      
#ifdef ATTACKTRACE
      {
	ostrstream tmp;
	tmp <<"attackeval-> simTime:" << simTime 
	    << " shortestTime: "<< shortestTime
	    << " distTime: " << distanceTime
	    << char(0);
	char* str= tmp.str();
	ATTACKTRC(MLog(str));
	delete str;
      }
#endif
      
      // at this point we have the point where the opponent will intercept 
      // the ball
      // simBall, and the time until interception simTime.

      // are we quiet sure to be able to intercept the ball before the opponent?
      interceptParams = skills->controlBall( availableStamina );
      if ( ( interceptParams.time < 3 && 
	     interceptParams.time + 3 < shortestTime ) ||
	   world->my.inCorridor( simBall.pos, world->ball.pos, 
				 world->my.kickRange() ))
	{
	  if ( world->my.inCorridor( simBall.pos, world->ball.pos, 
				     world->my.kickRange() ) ) 
	       ATTACKLOG("plAttack-> in corridor between ball & opp ");
	       
	  ATTACKLOG("plAttack-> intercepting ball ");
	  CalculatedCommand = true;
	  return true;
	}	  

      // first, handle offenders
      // offenders go to the position where the opponent will intercept the ball.
      // -> the opponent will be attacked shortly after he got the ball
      if ( tact->ourTeamOffMidfield( me.number ) || 
	   tact->ourTeamOffender( me.number ) )
	{
	  if ( tact->firstResponsiblePlayer == me.number && simTime != 0 )
	    {
	      interceptParams = skills->controlBall( availableStamina );
	      return true;
	    }
	  // don't do anything if a teammate is closer
	  if ( world->ourTeamClosestTo( simBall.pos ) != me.number )
	    return false;

	  interceptParams = 
	    skills->goToPositionTimed( simBall.pos, availableStamina, 
				       2*PLAYER_SIZE );
	  if (interceptParams.state == AdvSkills::TFINISHED )
	    {
	      interceptParams = 
		skills->goToPositionTimed( simBall.pos, availableStamina, 
					   1.0*PLAYER_SIZE );
	    };
	  ATTACKLOG("plAttack-> going to opponent ball intercept point ");
	  CalculatedCommand = true;
	  return true;
	};
	
      
      // Now try to simulate what the opponent is doing after interception.

      long FirstRespTime=0;
      long SecondRespTime=0;
      long ThirdRespTime=0;
      
      Vector2d Back = Vector2d(-1,0);
      Vector2d Opp2Goal = Vector2d(- FIELD_LENGTH / 2.0, 0) - simBall.pos;
      Vector2d OppPass = Vector2d(-FIELD_LENGTH/2.0+PENALTY_AREA_LENGTH /2.0,0)
	- simBall.pos;

      //Opp2Goal /= norm(Opp2Goal);
      // the further the opponent is back, the greater is the possibility
      // that he will go to our goal
      double tmp1=Step(-.5*FIELD_LENGTH+PENALTY_AREA_LENGTH-5,
		       -.5*FIELD_LENGTH+PENALTY_AREA_LENGTH+10,
		       simBall.pos.getX());
      Back=(tmp1*Opp2Goal+(1-tmp1)*Vector2d(-1,0));

      // if he is back and to the left or right of our goal he will
      // probably try to pass
      tmp1=Step(-.5*FIELD_LENGTH+PENALTY_AREA_LENGTH-5,
		-.5*FIELD_LENGTH+PENALTY_AREA_LENGTH,
		simBall.pos.getX())*
	(1 - Step(FIELD_WIDTH/6.0,PENALTY_AREA_WIDTH/2.0, 
		  fabs(simBall.pos.getY())));
      Back=tmp1*OppPass+ (1-tmp1) * Back;
      Back/=Back.getLength();
     
      interceptTime = 0;
      interceptPos  = OppAtBallInterception;
      bool finished = false;
      while ( interceptTime < 40 && !finished &&
	     interceptPos.getX() > - FIELD_LENGTH / 2.0 )
	{
	  interceptPos += Back * opp.vmax;
	  interceptTime++;

	  if ( norm( interceptPos - me.pos ) / (1.0*me.vmax) < 
	       interceptTime + simTime )
	    finished = true;
	  
	  //check responsible players
	  const WPlayer * our;
	  our = &world->ourTeam[tact->firstResponsiblePlayer];
	  if ( FirstRespTime ==0 && 
	       ( norm( interceptPos - our->pos ) / (1.0*our->vmax) < 
		 interceptTime + simTime ) )
		FirstRespTime=interceptTime;
	  
	  our = &world->ourTeam[tact->secondResponsiblePlayer];
	  if ( SecondRespTime ==0 && 
	       ( norm( interceptPos - our->pos ) / (1.0*our->vmax) < 
		 interceptTime + simTime ) )
	    SecondRespTime=interceptTime;
	  
	  our = &world->ourTeam[tact->thirdResponsiblePlayer];
	  if ( ThirdRespTime ==0 && 
	       ( norm( interceptPos - our->pos ) / (1.0*our->vmax) < 
		 interceptTime + simTime ) )
	    ThirdRespTime=interceptTime;
	}

#ifdef ATTACKTRACE
      ostrstream tmp;
      ATTACKTRC(LINE(simBall.pos,3*Back+simBall.pos));
      ATTACKTRC(CIRCLE(simOpp.pos,PLAYER_SIZE));
      ATTACKTRC(CIRCLE(simOpp.pos,PLAYER_SIZE-0.1));
      ATTACKTRC(CIRCLE(interceptPos,PLAYER_SIZE));
      ATTACKTRC(CIRCLE(interceptPos,PLAYER_SIZE+0.1));
      tmp <<"attackeval-> my interceptTime: "<< interceptTime
	  << " resp.Players' times: " << FirstRespTime << ", " << SecondRespTime << ", " << ThirdRespTime
	  << char(0);
      char* str= tmp.str();
      ATTACKTRC(MLog(str));
      delete str;
#endif
      CalculatedCommand = false;
  
      const int edge = 0;
      if ( ( tact->firstResponsiblePlayer != world->ourGoalieNo && 
	     FirstRespTime  !=0 && interceptTime > FirstRespTime + edge) ||
	   ( tact->secondResponsiblePlayer != world->ourGoalieNo && 
	     SecondRespTime !=0 && interceptTime > SecondRespTime + edge) || 
	   ( tact->thirdResponsiblePlayer != world->ourGoalieNo && 
	     ThirdRespTime  !=0 && interceptTime > ThirdRespTime + edge) )
	{
	  // a teammate is in a better position
	  if ( world->ourTeamClosestTo( opp.pos ) != me.number )
	    {
	      ATTACKTRC(MLog("evalplAttack-> teammate in better position"));
	      return false;
	    }
	  else
	    ATTACKTRC(MLog("evalplAttack-> teammate in better position, but I'm closest player"));

	};

      if ( me.number != tact->firstResponsiblePlayer &&
	   me.inCorridor( OppAtBallInterception,
			  OppAtBallInterception + 30 * opp.vmax * Back, 
			  2*me.kickRange() + PLAYER_SIZE, M_PI/6.0) &&
	   FirstRespTime !=0 )
	{
	  // I will be able to intercept opp and teammate is responsible.
	  // Don't move!
	  ATTACKTRC(MLog("evalplAttack-> will be able to intercept later"));
	  return false;
	};

      if (finished)
	{
	  // we have found an approximate position where we will be able 
	  // to intercept the opponent. 
	  // Are we already in his way to our goal?
	  ATTACKTRC(CORRIDOR( OppAtBallInterception,
			      OppAtBallInterception + 5 * opp.vmax * Back,
			      me.kickRange() + PLAYER_SIZE, M_PI/6.0));
	  if ( me.inCorridor( OppAtBallInterception,
			      OppAtBallInterception + 5 * opp.vmax * Back, 
			      me.kickRange() + PLAYER_SIZE, M_PI/6.0) )
	    {
	      ATTACKTRC(MLog("evalplAttack-> already in corridor, go to opp"));
	      interceptParams = 
		skills->goToPositionTimed( simBall.pos, availableStamina, 
					   1.5*PLAYER_SIZE );
	      if (interceptParams.state == AdvSkills::TFINISHED )
		{
		  interceptParams = 
		    skills->goToPositionTimed( simBall.pos, availableStamina, 
					       0.5*PLAYER_SIZE );
		};
	      CalculatedCommand = true;
	      return true;
	    };

	  // What will goToPos report if we just want to go to the interception
	  // point?
	  // first, try one step earlier  
	  interceptPos -= Back * opp.vmax;
	  interceptTime--;
	  
	  int t=0;
	  while (t<8)
	    {
	      interceptParams = 
		skills->goToPositionTimed( interceptPos, availableStamina, 
				       2*PLAYER_SIZE );
	      if ( interceptParams.time <= interceptTime + simTime )
		{
		  if (interceptParams.state == AdvSkills::TFINISHED )
		    {
		      interceptParams = 
			skills->goToPositionTimed( OppAtBallInterception, 
						   availableStamina, 
						   0.5*PLAYER_SIZE );
		    };
		  
#ifdef ATTACKTRACE
		  ostrstream tmp;
		  tmp <<"evalplAttack-> found intercept pos-1+"<< t << char(0);
		  char *str =tmp.str();
		  ATTACKTRC(MLog(str));
		  delete str;
#endif
		  CalculatedCommand = true;
		  return true;
		}
	      interceptPos += Back * opp.vmax;
	      interceptTime++;
	      t++;
	    };

	  double maxAngle = (interceptPos - me.pos).getAngle();
	  if ( maxAngle < 0 ) 
	    maxAngle+= 2 * M_PI;
	  double OptAngle = (interceptPos + Back * opp.vmax *3 - 
			    me.pos).getAngle();
	  if ( OptAngle < 0 ) 
	    OptAngle+= 2 * M_PI;
	  double BackAngle = Back.getAngle();
	  if ( BackAngle < 0 ) 
	    BackAngle+= 2 * M_PI;

	  Vector2d MyBack = Vector2d(-1,0);
	  Vector2d Me2Goal = Vector2d(- FIELD_LENGTH / 2.0, 0) - me.pos;
	  double tmp1=Step(-.5*FIELD_LENGTH+5,
			   -.5*FIELD_LENGTH+PENALTY_AREA_LENGTH,
			   me.pos.getX());
	  MyBack=(tmp1*Me2Goal+(1-tmp1)*Vector2d(-1,0));
	  
	  if ( maxAngle > BackAngle )
	    interceptParams = skills->runInDir(maxAngle,BackAngle,
					       OptAngle,availableStamina);
	  else interceptParams = skills->runInDir(BackAngle,maxAngle,
						  OptAngle,availableStamina);
	  CalculatedCommand = true;	  
	  
#ifdef ATTACKTRACE
	  ostrstream tmp;
	  tmp <<"attackeval-> didn't find goToParams interceptTime: "
	      << interceptTime  
	      << " params.Time " << interceptParams.time
	      << char(0);
	  char* str= tmp.str();
	  ATTACKTRC(MLog(str));
	  delete str;
#endif
	  return true;
	}
      else 
	{
	  ATTACKTRC(MLog("evalplAttack-> no intercept point found,"
			 " going back"));

	  Vector2d MyBack = Vector2d(-1,0);
	  Vector2d Me2Goal = Vector2d(- FIELD_LENGTH / 2.0, 0) - me.pos;
	  double tmp1=Step(-.5*FIELD_LENGTH+5,
			   -.5*FIELD_LENGTH+PENALTY_AREA_LENGTH,
			   me.pos.getX());
	  MyBack=(tmp1*Me2Goal+(1-tmp1)*Vector2d(-1,0));
	  
	  
	  // if the opponent is on our outside, go to our goal else
	  // go straight back
	  if ( me.pos.getY() * simBall.pos.getY() > 0 &&
	       fabs( me.pos.getY() ) - fabs( simBall.pos.getY() ) > 0 )
	    {
	      interceptParams = 
		skills->goToPositionTimed( me.pos + 10 * Vector2d(-1,0), 
					   availableStamina, 
					   3*PLAYER_SIZE );
	    }
	  else
	    {
	      interceptParams = 
		skills->goToPositionTimed( me.pos + 10 * MyBack, 
					   availableStamina, 
					   3*PLAYER_SIZE );
	    };

	  //skills->runInDir(-M_PI+.001,M_PI-.001);
	  CalculatedCommand = true;
	  return true;
	}
    }

  return false;
}

void
PositionModule::playerAttack(int OpponentNo) {
  
  if (OpponentNo == UNKNOWN) {
    CERR ("act -> attack failed, because OpponentNo is not set ! Intercepting instead.");
    AdvSkills::TurtleCommand   turtleParams =skills->controlBall( availableStamina );
    skills->exec(turtleParams);
    return;
  }
 
  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! TODO
  //- player intercept (simBall) out of field?!?
  //- uses too much stamina?

  AdvSkills::TurtleCommand   turtleParams;  
  WPlayer opponent = simOpp;
  
  // ball is not kickable right now (otherwise we wouldnt attack, but dribble)

  double stam    = availableStamina;

#ifdef ATTACKTRACE
  {
  ostrstream tmp;
  ATTACKTRC(CIRCLE(interceptPos,PLAYER_SIZE));
  ATTACKTRC(CIRCLE(interceptPos,PLAYER_SIZE+0.1));
  tmp <<"attackeval-> final interceptTime: "<< interceptTime  << char(0);
  char* str= tmp.str();
  ATTACKTRC(MLog(str));
  delete str;
  };
#endif

#ifdef ATTACKTRACE
  ostrstream tmp;
  ATTACKTRC(CIRCLE(simBall.pos,BALL_SIZE));
  ATTACKTRC(CIRCLE(world->ball.pos,world->my.kickRange()));
  tmp <<"attack-> attacking player no. : "<< OpponentNo<< "("<< simTime << ")"
      << " lastAttack: " << (lastAttack==GO_BACK ? "Go Back" :
			     lastAttack==PLAYER_INTERCEPT ? "intercept Pl ":
			     "unknown") << char(0);
  char* str= tmp.str();
  ATTACKTRC(MLog(str));
  delete str;
#endif

  AdvSkills::TurtleCommand goToParams = 
    skills->goToPositionTimed( simBall.pos, stam, world->my.kickRange() );

  if( simTime <= 1 )  // i.e. ball is close to opponent's kickrange
    {

      //does it make sense to tackle?
      AdvSkills::TackleCommand tcom = skills->tackle();

      // when do we want to tackle? (be more careful in defensive situation)
      double useFailProb = ( me.pos.getX() < 0.0 ? 0.2 : 0.3 );
      if ( tact->ourTeamDefender( me.number ) )
	useFailProb /= 2.0;
      
      if (tcom.failProb < useFailProb &&
	  tcom.state != AdvSkills::TKICKABLE && 
	  tcom.state != AdvSkills::TRANGE &&
	  world->ball.age == 0 )
	{
	  bool doit = false;   // really tackle?

	  // if in our half tackle to the outside of the field, else to the inside
	  if( me.pos.getX() < 0.0 )
	    {
	      if( fabs( me.bodyDir ) < M_PI_4 ||
		  ( me.pos.getY() < 0.0 && me.bodyDir > -3.0 * M_PI_4 && me.bodyDir <= 0.0 ) ||
		  ( me.pos.getY() > 0.0 && me.bodyDir <  3.0 * M_PI_4 && me.bodyDir >= 0.0 ) )
		doit = true;
	    }
	  else //if( ( fabs( me.pos.getY() ) < FIELD_WIDTH / 4.0 && fabs( me.bodyDir ) < M_PI_2 ) ||
	    //   fabs( me.bodyDir ) < M_PI_4 ||   
	    //   ( me.pos.getY() > 0.0 && me.bodyDir > -M_PI_2 && me.bodyDir <= 0.0 ) ||
	    //   ( me.pos.getY() < 0.0 && me.bodyDir <  M_PI_2 && me.bodyDir >= 0.0 ) )
	    if ( fabs( me.bodyDir - 
		       ( Vector2d( FIELD_LENGTH / 2.0 - PENALTY_AREA_LENGTH / 2.0, 0.0 ) - me.pos ).getAngle() )
		 < M_PI_4 )
	    doit = true;

	  if( doit )
	    {
	      skills->exec(tcom);
	      tact->reportOurTacticalAction(TA_REACT, me.number);
	      action=TACKLE;
	      
#ifdef ATTACKTRACE
	      {
		ostrstream tmp;
		tmp <<"attack-> tackling opponentNo:  "<< OpponentNo
		    <<" bodyDir: " << me.bodyDir << " me.pos: " << me.pos
		    <<" failProb: " << tcom.failProb
		    <<" me2goal: " 
		    << ( Vector2d( FIELD_LENGTH / 2.0 - PENALTY_AREA_LENGTH / 2.0, 0.0 ) - me.pos ).getAngle()
		    << char(0);
		char* str= tmp.str();
		ATTACKTRC(MLog(str));
		delete str;
	      }
#endif
	      CERR(" Tackled. ");
	      return;
	    }
	  else 
	    {
	      ATTACKTRC(MLog("attack-> tackle possible but bodyDir wrong"));
	      CERR("Didn't tackle because of wrong bodyDir: " << me.bodyDir);
	    };
	}
      else 
	{
#ifdef ATTACKTRACE
	  {
	    ostrstream tmp;
	    tmp <<"attack-> tackle success unlikely: OppNo:  "<< OpponentNo
		<<" bodyDir: " << me.bodyDir << " me.pos: " << me.pos
		<<" failProb: " << tcom.failProb
		<< char(0);
	    char* str= tmp.str();
	    ATTACKTRC(MLog(str));
	    delete str;
	  }
#endif
	}
    }
  else if (simTime == 0) // opponent is in kick range
    {        
      // if opponent is not in my way, try to intercept directly
      // !! abwehrspieler sollten sich nicht komplett um den Spieler ziehen 
      // lassen
      
      if( goToParams.time <= 1 &&
	  world->my.distTo(simBall.pos) < world->my.distTo(opponent)) {
	
	if( goToParams.state == AdvSkills::TFINISHED )
	  ATTACKTRC(MLog("attack->skills finished :__LINE__"));

	skills->exec( goToParams );
	ATTACKTRC(MLog("attack-> going directiy to ball, exec"));
	return;
      }
    }
    
  if (CalculatedCommand)
    {
      // executing previously calculated command
      skills->exec(interceptParams);

      WBall nextBall = world->ball;
      // if opp is not in kickrange
      if ( simTime !=0 ) 
	nextBall.simulate();

      // calculate a turnHeadWish
      view_width vWidth;
      // Angle width;   // unused! [CM]
      Angle myNextBodyDir = interceptParams.bodyDir;
      Angle oppAngle = ( simOpp.pos - me.pos ).getAngle() - myNextBodyDir;
      Angle ballAngle = ( nextBall.pos - me.pos ).getAngle() - myNextBodyDir;

      if ( simOpp.age > 4 || simOpp.coneTime > 0 )
	{
	  // we have to look at opponent !!!
	  // find out smallest possible view angle
	  if ( simOpp.coneTime == 2 )
	    vWidth = VIEW_WIDTH_WIDE;
	  else if ( simOpp.coneTime == 1 )
	    vWidth = VIEW_WIDTH_NORMAL;
	  else vWidth = VIEW_WIDTH_NARROW;
	  
	  // is it possible to see opp?
	  while ( fabs( oppAngle ) > 
		  fabs( DEG_TO_RAD( MAXNECKANG ) + world->getViewAngle( vWidth ) ) )
	    vWidth = ( vWidth == VIEW_WIDTH_NARROW ? VIEW_WIDTH_NORMAL :
		       vWidth == VIEW_WIDTH_NORMAL ? VIEW_WIDTH_WIDE :
		       VIEW_WIDTH_WIDE );

	  ATTACKTRC(MLog("attack -> opp too old, looking for him"));
	  plAttackThw.lookAtPos( Grade::PERFECT, simOpp.pos, vWidth);
	  return;
	}

      if ( world->ball.age > 4 || world->ball.coneTime > 0 )
	{
	  // we have to look at ball !!!
	  // find out smallest possible view angle
	  if ( world->ball.coneTime == 2 )
	    vWidth = VIEW_WIDTH_WIDE;
	  else if ( world->ball.coneTime == 1 )
	    vWidth = VIEW_WIDTH_NORMAL;
	  else vWidth = VIEW_WIDTH_NARROW;
	  
	  // is it possible to see the ball?
	  while ( fabs( ballAngle ) > 
		  fabs( DEG_TO_RAD( MAXNECKANG ) + world->getViewAngle( vWidth ) ) )
	    vWidth = ( vWidth == VIEW_WIDTH_NARROW ? VIEW_WIDTH_NORMAL :
		       vWidth == VIEW_WIDTH_NORMAL ? VIEW_WIDTH_WIDE :
		       VIEW_WIDTH_WIDE );

	  ATTACKTRC(MLog("attack -> ball too old, looking for it"));
	  plAttackThw.lookAtPos( Grade::PERFECT, nextBall.pos, vWidth);
	  return;
	};

      // ok ball and opponent are both not too old,
      // what can we do?
      if ( shortestTime < 3 )
	vWidth = VIEW_WIDTH_NARROW;
      else if ( shortestTime < 6 && me.distTo(simOpp) < 4 * PLAYER_SPEED_MAX )
	vWidth = VIEW_WIDTH_NORMAL;
      else vWidth = VIEW_WIDTH_WIDE;

      // if both opp and ball are young and opp is not in kickrange,
      // try to have a look around
      if ( simOpp.age < 2 && world->ball.age < 2 &&
	   simTime > 1 )
	{
	  plAttackThw.lookAudio( Grade::PERFECT, vWidth );
	  ATTACKTRC(MLog("attack -> lookAudio"));
	  return;
	}
      
      // what vWidth do I need to look at the older object?
      Angle oldAngle = ( world->ball.age > simOpp.age ? ballAngle : oppAngle );
      while ( fabs( oldAngle ) > 
	      fabs( DEG_TO_RAD( MAXNECKANG )  + world->getViewAngle( vWidth ) ) )
	vWidth = ( vWidth == VIEW_WIDTH_NARROW ? VIEW_WIDTH_NORMAL :
		   vWidth == VIEW_WIDTH_NORMAL ? VIEW_WIDTH_WIDE :
		   VIEW_WIDTH_WIDE );

      // is it possible to see both opp and ball?
      if ( fabs( ballAngle - oppAngle ) < 1.8 * world->getViewAngle( vWidth ) &&
	   fabs( ballAngle + oppAngle ) / 2.0 < 
	   fabs( DEG_TO_RAD( MAXNECKANG )  + world->getViewAngle( vWidth ) ) )
	{
	  ATTACKTRC(MLog("attack -> trying to see both ball and opp"));
	  ATTACKTRC(LINE(me.pos,me.pos+radAng2Vector2d( 10.0,
				 Angle( 0.5 * ( ballAngle + oppAngle )))));
	  plAttackThw.lookAtPos( Grade::PERFECT,
				 me.pos + radAng2Vector2d( 10.0, myNextBodyDir +
				 Angle( 0.5 * ( ballAngle + oppAngle ) ) ),
				 vWidth);
	  return;
	}

      ATTACKTRC(MLog("attack -> turning head to older object"));
      plAttackThw.lookAtPos( Grade::PERFECT,
			     me.pos + radAng2Vector2d( 10.0, myNextBodyDir + 
						       oldAngle ),
			     vWidth);
      return;
    };

  Vector2d target;
  
  // if I am between opponent and ball, just intercept
  if( world->my.inCorridor( opponent.pos, world->ball.pos, world->my.kickRange() ) )
    {
      ATTACKTRC(MLog("attack -> between ball and opp., intercepting ball"));
      turtleParams = skills->controlBall( stam );
      
      if( turtleParams.state == AdvSkills::TFINISHED )
	ATTACKTRC(MLog("attack->skills finished in line "));
      
      skills->exec( turtleParams );
      return;
    }

  //  const double mindist=6*PLAYER_SIZE;
  // is opponent closer to our goal? 
  if  ( fabs( ( me.pos - opponent.pos ).getAngle() ) < M_PI / 3.5 +
	Triangle(M_PI / 6.0, FIELD_WIDTH / 2.0, me.pos.getY() )  ||
	 ( lastAttack == GO_BACK && 
	   ( fabs( ( me.pos - opponent.pos ).getAngle() ) < M_PI / 3.0 +
	     Triangle(M_PI / 6.0, FIELD_WIDTH / 2.0, me.pos.getY() ) ) )
	) 
    {  //ups, opponent is far behind, I have to think of something really clever!
     
        const double mindist=6*PLAYER_SIZE;
        ATTACKTRC(MLog("attack->opponent too far behind, going back"));
	lastAttack=GO_BACK;

        if (!((fabs(opponent.pos.getX())+mindist<me.pos.getX()) && 
              (fabs(opponent.pos.getX())-mindist>me.pos.getX())))
          {
            target.setY(me.pos.getY());
            ATTACKTRC(MLog("attack-> going straight back"));
          }
        else 
          {
            target.setY(opponent.pos.getY()<0?opponent.pos.getY()
                        +mindist:opponent.pos.getY()-mindist);
            ATTACKTRC(MLog("attack-> trying to get between center and opponent"));
          };
        target.setX(opponent.pos.getX()-6*mindist);
        //hier auch simBall oder opponent
        ATTACKTRC(LINE(me.pos,target));
	goToParams   = skills->goToPositionTimed(target,stam,
						 1.0*PLAYER_SIZE);
	ATTACKTRC(MLog("attack-> executing skills early"));
	skills->exec(goToParams);
	return;
      
      /*
      if  ( ( fabs( me.pos.getY() ) <  fabs( opponent.pos.getY() ) )  && 
	    ( me.pos.getY()*opponent.pos.getY()>0) )  
	{ //ich wei nicht ob hier nicht statt simBall opponent besser wre, 
	  //ich habe das mal so gehabt und dann spter gendert
	  ATTACKTRC(MLog("attack->opponent too far behind, but I'm in the middle,going back"));
	  
	  target.setY(me.pos.getY());
	  
	  target.setX(opponent.pos.getX()-mindist);
	  //hier auch simBall oder opponent
	  ATTACKTRC(LINE(me.pos,target));  
	  goToParams   = skills->goToPositionTimed(target,stam,
						   1.0*PLAYER_SIZE);
	  ATTACKTRC(MLog("attack-> executing skills early"));
	  skills->exec(goToParams);
	  return;
	  
	}
      else
	{
	  //hmm, opponent is behind and on the inside. He better makes some kind of mistake
	  //or I won't have a chance to get him
	  ATTACKTRC(MLog("attack->opponent too far behind and in the middle, going back!"));
	  
	  target.setY(opponent.pos.getY()<0?opponent.pos.getY()-mindist:opponent.pos.getY()+mindist);
	  target.setX(opponent.pos.getX()-(me.pos.getX()-opponent.pos.getX()));

	  //hier auch simBall oder opponent
	  ATTACKTRC(LINE(me.pos,target));  
	  goToParams   = skills->goToPositionTimed(target,stam,
						   1.0*PLAYER_SIZE);
	  ATTACKTRC(MLog("attack-> executing skills early"));
	  skills->exec(goToParams);
	  return;
	  
	}
      */
    }
  else
    if ((simBall.pos.getX() > 0) &&
	(!tact->ourTeamDefender(me.number))) 
      //d.h. ball ist in gegn. Hlfte
      { 
	// it's likely the ball is kicked, so attack ball directly
	target = simBall.pos;
      }
    else 
      { //ball in unserer Hlfte
	// goto the position where the opponent is trying to
	// intercept ball, but stay on the side of our goal
	lastAttack=PLAYER_INTERCEPT;

	if (((world->ball.pos-opponent.pos).getLength()<world->my.kickRange()*1.1) ||
	    ((simTime==0) && (opponent.age<=1)))
	  {
	    ATTACKTRC(MLog("attack-> opponent in kickrange, intercepting opp."
			   ));
	    target=opponent.pos;
	  }
	else target=simBall.pos;
	
	turtleParams   = 
	  skills->goToPositionTimed(target, 
				    world->my.stamina, PLAYER_SIZE);
	int reachtime=turtleParams.time;
	
#ifdef ATTACKTRACE
	ostrstream  tmp;
	tmp  << "attack-> trying to intercept player, time needed to "<<
	  "get to opp. interc. ball point: "<< 
	  reachtime<< char(0);
	{ char * str= tmp.str();
	ATTACKTRC(MLog(str));
	delete str; }
#endif
	
	Vector2d InterceptDir=(Vector2d(-0.5*FIELD_LENGTH+PENALTY_AREA_LENGTH / 3.0,0)- simBall.pos);
	InterceptDir=1/InterceptDir.getLength()*InterceptDir;
	double tmp1=Step(-.5*FIELD_LENGTH+PENALTY_AREA_LENGTH+5,
			 -.5*FIELD_LENGTH+PENALTY_AREA_LENGTH+15,
			 simBall.pos.getX());

	InterceptDir=(tmp1*InterceptDir+(1-tmp1)*Vector2d(-1,0));
	InterceptDir=1/InterceptDir.getLength()*InterceptDir;
	//pretend that the opponent will go straight to our goal if he is
	//next to our penalty area, otherwise pretend he will go straight back
	/*
	if ((simTime<3) && 
	    ((goalShotEval->otherEval(world, target)
	      ).is_better_than(Grade::AVERAGE)))
	{//immediate danger, the opponent will propably shoot into our goal, so
	  // forget everything and try to get between opponent and goal.
	  InterceptDir*=skills->kickRange+PLAYER_SIZE;
	  stam=world->my.stamina;
	  ATTACKTRC(MLog("attack-> opponent will propably shoot into our goal!!"));
	}
	else*/
	{
	  const double OppMaxSpeed=0.9;
	  Vector2d Me2Opp=1/PLAYER_SPEED_MAX*(target-me.pos);
	  
	  tmp1=(-simTime+InterceptDir*Me2Opp*OppMaxSpeed)/
	    (1-OppMaxSpeed*OppMaxSpeed);
	  double tmp2=(-sqrnorm(Me2Opp)+simTime*simTime)/
	    (1-OppMaxSpeed*OppMaxSpeed);
	  if (tmp1*tmp1<tmp2) 
	    {
	      ATTACKTRC(MLog("attack-> oh shit !! negativ value"));
	      CERR("Wurzel negativ!!!");
	      target=simBall.pos;
	      target.setY(me.pos.getY()+10);
	    }
	  else 
	    {//calculation of player intercepttime seems to be ok
#ifdef ATTACKTRACE
	      ostrstream temp1;
	      temp1 <<"attack->  c^2: "<< sqrnorm(Me2Opp)<< "tmp1: "
		    << tmp1<< " tmp2 : " <<tmp2 <<" b*C"<< InterceptDir*Me2Opp
		    <<char(0);
	      { char * str = temp1.str();
	      ATTACKTRC(MLog(str));
	      delete str; }
#endif
	      tmp1+=sqrt((tmp1*tmp1)-tmp2);
	    
	      turtleParams   = 
		skills->goToPositionTimed(tact->ourTeamReachBallPos[me.number], 
					  stam, PLAYER_SIZE);
	      
	      if ((turtleParams.time<simTime-1) || (reachtime<simTime-1))
		//if we can reach the ball before opponent, try to do it
		{
		  ATTACKTRC(MLog("attack->turtleParams.time<simTime!Ballhandling?"));
		  // if (turtleParams.time<reachtime)
		    {
		      ATTACKTRC(MLog("attack-> executing control ball"));
		      //ATTACKTRC(LINE(me.pos,tact->ourTeamReachBallPos[me.number]));
		      //ATTACKTRC(POSITION(tact->ourTeamReachBallPos[me.number]));
		      
		      turtleParams = skills->controlBall( stam );

		      if( turtleParams.state == AdvSkills::TFINISHED )
			ATTACKTRC(MLog("attack->skills finished in PosFKT.C line: __LINE__"));

		      skills->exec(turtleParams);
		      return;
		    }
		  /*	  else
		    {
		      turtleParams = skills->controlBall( stam );
		      //   if ( turtleParams.time < reachtime )
			{
			  ATTACKTRC(MLog("attack-> executing control ball"));
			  skills->exec(turtleParams);
			  return;
			}

		      InterceptDir=Vector2d(0,0);
		      turtleParams =
			skills->goToPositionTimed(simBall.pos, 
						  stam, PLAYER_SIZE);
		      if( turtleParams.state == AdvSkills::TFINISHED )
			ATTACKTRC(MLog("attack->skills finished in PosFKT.C line: ""__LINE__"));

		      skills->exec(turtleParams);
		      return;
		      };*/		
		}
	      else
		{
		  if (tmp1<0)
		    {
		      ATTACKTRC(MLog("attack -> tmp1 negativ doing safty hack"));
		      CERR("Mist! negativer Wert. das sollte nichtmehr vorkommen");
		      tmp1=fabs(tmp1);
		    }
		};
		    
#ifdef ATTACKTRACE
	      ostrstream temp;
	      temp <<"attack-> time to intercept after opp. got ball: "
		   << tmp1<< char(0);
	      char * str = temp.str();
	      ATTACKTRC(MLog(str));
	      delete str;
#endif	  
	      
	    }; // if tmp1*tmp1<tmp2 else	  
	
	  InterceptDir*=.5*world->my.kickRange()+tmp1*OppMaxSpeed*PLAYER_SPEED_MAX;
	};	  
	ATTACKTRC(LINE(target,target+InterceptDir));
	target=target+InterceptDir;

	goToParams   = skills->goToPositionTimed(target,stam,
						 1.5*PLAYER_SIZE);
	if ( ( simTime==0 ) &&
	     ( fabs( ( me.pos - opponent.pos ).getAngle() ) < 3.0 / 4.0 * M_PI ) )
	  {
	    // opponent is in kickrange and I'm more or less behind him.
	    // He will almost surely kick the ball again to our goal, so don't turn
	    // when we have been going back
	    ATTACKLOG("plAttack-> opponent behind and in kickrange");
	    if ( ( goToParams.command==AdvSkills::TurtleCommand::TURN ) &&
		 ( fabs( me.bodyDir ) > 4.0 / 5.0 * M_PI ) )
	      {
		goToParams   = skills->goToPositionTimed(me.pos+radAng2Vector2d(2,me.bodyDir),
							 stam,1.5*PLAYER_SIZE);
	        ATTACKLOG("plAttack-> don't turn now");
	      }
	    else ATTACKLOG("plAttack-> goToParams ok");
	    skills->exec(goToParams);
	    return;
	  };

      } // else..  ball in our half 
    ATTACKTRC(POSITION(target));
    //target=goAroundOpponent(target,opponent,distToOpponent,1);
    goToParams   = skills->goToPositionTimed(target,stam,
					     1.5*PLAYER_SIZE);
#ifdef ATTACKTRACE
    {
      ostrstream tmp;
      tmp << "attack-> used Stamina: " << stam<< char(0);
      { char * str = tmp.str();
      ATTACKTRC(MLog(str));
      delete str; }
    };
#endif

    //    goToParams = goAroundOpponent(target, opponent, distToOpponent, 
    //				  world->my.stamina);
    // ist das nicht zuviel stamina?!?
    //ATTACKTRC(MLog("attack -> first attacker; calculated goAroundOpponent"));
 
  // goToParams from one of the goAroundOpponent() above
  if( goToParams.state == AdvSkills::TFINISHED )
    {
      ATTACKTRC(MLog("attack -> reached target, intercepting ball"));
      goToParams = skills->goToPositionTimed( world->ball.pos, stam, 
					      world->my.kickRange() );
      if( goToParams.state == AdvSkills::TFINISHED )
	{
	  ATTACKTRC(MLog("attack -> ??? already reached ball ???"));
	  CERR("attack-> already reached the ball?? What about ballhandling?");
	  skills->noAction();
	  return;
	}
      skills->exec( goToParams );
      return;
    }
  
  if( goToParams.state == AdvSkills::TFINISHED )
    ATTACKTRC(MLog("attack->skills finished in PosFKT.C line: __LINE__"));
  
  ATTACKTRC(MLog("attack -> exec skills fallthrou"));
  skills->exec( goToParams );
  return;
}


Vector2d 
PositionModule::goAroundOpponent(Vector2d target,
				 WPlayer opponent,
				 double dist,double inside) {
  //inside decides if a player decides whether a player tries to go 
  //around the opponent on the outside or on the inside of the field
  //normally it should be (me.pos-target).getX(), but for attacking 
  //opponents, it is better to always go on the inside (the side facing
  //to our goal). Inside has to be positive in this case.

  Vector2d newTarget;
  double mindist=(opponent.pos-me.pos).getLength();
  mindist=mindist<4*PLAYER_SIZE?mindist:4*PLAYER_SIZE;

  ATTACKTRC(MLog("Attack-> goAroundOpponent"));

  //nchst Annherung an den Gegner, wenn ich geradeaus laufe
  Vector2d Odist=(target-me.pos);
  Odist=(me.pos+(1/(Odist.getLength()*Odist.getLength())*
		 Odist*(opponent.pos-me.pos))*Odist)-opponent.pos;

  //wenn die kleiner ist als mindist 
  //und der Gegner zwischen mir und meinem Ziel ist mu ich was tun!
  if ((Odist.getLength()<mindist) && 
      ((me.pos-target).getLength()>(me.pos-opponent.pos).getLength()) && 
      ((Odist+opponent.pos-target).getLength()
       <(me.pos-target).getLength()))
    {
      ATTACKTRC(MLog("Attack-> goAroundOpponent: in critical distance"));
      ATTACKTRC(LINE(target,me.pos));
      ATTACKTRC(LINE(opponent.pos,opponent.pos+Odist));
      
      //je nach der Variablen inside auf der inneren Seite vorbeilaufen,
      //oder auen
      Odist=target-me.pos;
      Odist*=1/Odist.getLength();
      Odist=mindist*Odist.perp();
      if (inside*Odist*(me.pos-Vector2d(-FIELD_LENGTH/2.0,0))<0)
	newTarget=opponent.pos+Odist;
      else newTarget=opponent.pos-Odist;

      Odist=newTarget-me.pos;
      Odist=1/Odist.getLength()*Odist;
      newTarget=newTarget+4*Odist;
      ATTACKTRC(LINE(me.pos,newTarget));
      ATTACKTRC(POSITION(newTarget));
    }
  else newTarget=target;
  return newTarget; 
}

double PositionModule::GradeToDouble(Grade value) 
{
  if (value.is_like(Grade::PERFECT))
    return 5 ;
  if (value.is_like(Grade::GOOD))
    return 4 ;
  if (value.is_like(Grade::AVERAGE))
    return 3 ;
  if (value.is_like(Grade::POOR))
    return 2 ;
  if (value.is_like(Grade::BAD))
    return 1 ;
  if (value.is_like(Grade::IMPOSSIBLE))
    return 0 ;
  return 0; 
}

Grade PositionModule::DoubleToGrade(double value)
{
  if ((0<=value)&&(value<0.5))
    return Grade::IMPOSSIBLE;
  if ((0.5<=value)&&(value<1.5))
    return Grade::BAD;
  if ((1.5<=value)&&(value<2.5))
    return Grade::POOR;
  if ((2.5<=value)&&(value<3.5))
    return Grade::AVERAGE;
  if ((3.5<=value)&&(value<4.5))
    return Grade::GOOD;
  if ((4.5<=value)&&(value<6))
    return Grade::PERFECT;
  return Grade::IMPOSSIBLE; 
}

Vector2d PositionModule::gotoFreePos()
{
  /*
  double PlayerScore[12];   // scores for different players
  double BodyFactor;
  Vector2d homePos = positionEval->GetPlayerPosition(world,me.number);
  Vector2d ClosestApproach;
  const WPlayer &BPlayer = world->ourTeam[tact->firstResponsiblePlayer];
  Vector2d Pl2Me = me.pos - BPlayer.pos;
  Pl2Me /= norm( Pl2Me );
  Vector2d Pl2Home = homePos - BPlayer.pos;
  double Pl2HomeDist= norm( Pl2Home );
  Pl2Home /= Pl2HomeDist;
  
#ifdef FREEPOSTRACE
  ostrstream t;
#endif
  double duration;
  double distance;
  double ReachFactor;
  double minReachFactor = 99999;
  double minIndex = 999;
  Vector2d Pl2Pos;
  for (int j = -10 ; j <=10; j++)
    {
      ReachFactor = 0.0;
      Pl2Pos = me.pos + Vector2d(j,0) - BPlayer.pos;
      distance = norm( Pl2Pos ); 
      Pl2Pos /= distance;
      
#ifdef FREEPOSTRACE
      t << j << "(";
#endif
      // find opponents which could intercept a possible pass
      for( int i = 1; i <= 11; i++ )
	{
	  const WPlayer &opp = world->otherTeam[i];

	  BodyFactor = Pl2Pos * ( opp.pos - BPlayer.pos);   
	  // is the opponent in pass direction?
	  if ( BodyFactor < 0 )
	    continue;
   
	  ClosestApproach = BodyFactor *  Pl2Home;
	  
	  // opp is behind intercept partner
	  if ( norm( ClosestApproach ) > distance + 5 * opp.vmax  )
	    continue;
	  
	  duration = log( 1 - norm( ClosestApproach ) / ( 0.8 * BALL_SPEED_MAX ) 
			  * ( 1- BALL_DECAY ) ) / log( BALL_DECAY);
	  
	  ReachFactor += opp.vmax / 
	    (norm( BPlayer.pos + ClosestApproach - opp.pos ) / duration); 
#ifdef FREEPOSTRACE
	  t << i << ":" << ReachFactor << ", " << duration << " ";
#endif
	}
#ifdef FREEPOSTRACE
      t << ") ";
#endif
      if ( ReachFactor < minReachFactor )
	{
	  minReachFactor = ReachFactor;
	  minIndex = j;
	};
    };
  FREEPOSTRC(LINE(BPlayer.pos,me.pos+Vector2d(minIndex,0)));
      
#ifdef FREEPOSTRACE
  t << char(0);
  char *str=t.str();
  FREEPOSLOG("dangerous opps: " <<str);
  delete str;
#endif
  return homePos;
  */

    Worldmodel testworld = *world;
    WPlayer testme;
    int i,j,k,BestGradeNumber=0;
    Grade BestGrade;
    Vector2d checkPositions[NumberofPointsX*NumberofPointsY+1] ;
    Grade checkGrades[NumberofPointsX*NumberofPointsY+1] ;
    Grade middleGrades[NumberofPointsX*NumberofPointsY+1];
    Vector2d checkBestPosofPlayer[NumberofPointsX*NumberofPointsY+1] ;
    Vector2d middlePositionsCutOff[NumberofPointsX*NumberofPointsY+1] ;
    Vector2d min,max;
    double middleX, middleY;
    bool playerNear[NumberofPointsX*NumberofPointsY+1];
    int mindistNumber;
    double xFront, xBack, yFrontBack;
    positionEval->GetPlayerArea(world,me.number,min,max);
    FREEPOSTRC(BOX(min,max));
    //Where is best position to receive pass?
    //First check real position of player.
    checkPositions[0] = me.pos ;
    if (tact->firstResponsiblePlayer != UNKNOWN )
      checkGrades[0] = passEval->ourEval(world, tact, tact->firstResponsiblePlayer,
					 me.number, checkBestPosofPlayer[0]);
    else
      checkGrades[0].set_BAD();

    testme = me;
    //Where is best position for pass if player changes his position in the playerArea.
    //Divide PlayerArea into NumberofPointsX*NumberofPointsY  pieces and check the middle point of each piece:
    //     I   I
    //  -----------
    //     I   I 
    //  -----------
    //     I   I
   
    playerNear[0]=false ; //even if there's a player near your own position you should stay there if there's no better position
    //now check other positions in playerarea
    for (i=0;i<NumberofPointsX;i++)
    for (j=0;j<NumberofPointsY;j++)
      {
        //change PlayerPosition
        middleX=min.getX()+i*(max.getX()-min.getX())/NumberofPointsX+(max.getX()-min.getX())/(NumberofPointsX*2.);
        middleY=min.getY()+j*(max.getY()-min.getY())/NumberofPointsY+(max.getY()-min.getY())/(NumberofPointsY*2.);
	checkPositions[i*NumberofPointsY+j+1].setX(middleX);
	checkPositions[i*NumberofPointsY+j+1].setY(middleY);
	testme.pos=checkPositions[i*NumberofPointsY+j+1];			  
	testworld.ourTeam[me.number] = testme;
        //evaluate new Position
	if (tact->firstResponsiblePlayer != UNKNOWN )
	  checkGrades[i*NumberofPointsY+j+1] = 
	    passEval->ourEval(&testworld, tact, tact->firstResponsiblePlayer,
			      me.number, checkBestPosofPlayer[i*NumberofPointsY+j+1]);
	else
	  checkGrades[i*NumberofPointsY+j+1].set_BAD();
      }
       
    //if FreePos has not been active for a longer time start new calculation of middle points
    if(world->time- lastActive > NOT_ACTIVE_TOO_LONG)
      {
        for (i=0;i<(NumberofPointsX*NumberofPointsY+1); i++)
          {
            middleGrades[i].set_IMPOSSIBLE() ;
	    sizeOfMiddleGrades[i]=0;
	    middlePositions[i].setXY(1000,1000);
          }
      } 
    for (i=0; i<(NumberofPointsX*NumberofPointsY+1); i++)
      {
        //calculate middlePositions
        if (middlePositions[i].getX()==1000) //start new calculation 
          middlePositions[i]= checkBestPosofPlayer[i];
        else
          middlePositions[i]= (middlePositions[i]*sizeOfMiddleGrades[i]+checkBestPosofPlayer[i]*GradeToDouble(checkGrades[i]))*
          (1/(sizeOfMiddleGrades[i]+GradeToDouble(checkGrades[i])));
	if(sizeOfMiddleGrades[i]==0)
	  {
	    middleGrades[i] = checkGrades[i];
            sizeOfMiddleGrades[i]=GradeToDouble(checkGrades[i]);
	  }
        else
          {
            sizeOfMiddleGrades[i] =  (OLD_GRADE_FACTOR*sizeOfMiddleGrades[i]+NEW_GRADE_FACTOR*GradeToDouble(checkGrades[i]))
	      /(OLD_GRADE_FACTOR+NEW_GRADE_FACTOR);
            middleGrades[i] = DoubleToGrade(sizeOfMiddleGrades[i]) ;
	  }
#ifdef FREEPOSTRACE
	{
	  int t(0),x(0),y(0);		  
	  for (t=0; t<NumberofPointsX; t++)
	    if((NumberofPointsY*t<i)&&(i<(t+1)*NumberofPointsY+1))
	      x=t ;
	  y=i-(x*NumberofPointsY)-1;
	  ostrstream tmp;
	  tmp << "gotoFreePos -> Position (" << x << ", " << y << ")  middleGrade " << middleGrades[i] <<"checkGrade" <<checkGrades[i] << char(0) ;
	  char * str= tmp.str();
	  FREEPOSTRC(MLog(str));
	  delete str;
	  ostrstream tmp2;
	  tmp2 <<  "gotoFreePos -> Position (" << x << ", " << y << ")  SizeofMiddleGrade " << sizeOfMiddleGrades[i] <<char(0);
	  str = tmp2.str();
	  FREEPOSTRC(MLog(str));
	  delete str;
	}
#endif
        //cut off positions which are outside playerarea
        middlePositionsCutOff[i] = middlePositions[i] ;
        if (middlePositionsCutOff[i].getX() >max.getX())
          middlePositionsCutOff[i].setX(max.getX()) ;
	if (middlePositionsCutOff[i].getX() <min.getX())
          middlePositionsCutOff[i].setX(min.getX());
	if (middlePositionsCutOff[i].getY() >max.getY())
          middlePositionsCutOff[i].setY(max.getY()) ;
	if (middlePositionsCutOff[i].getY() <min.getY())
          middlePositionsCutOff[i].setY(min.getY()) ;
        //is one of our players near the middlePosition?
        playerNear[i]=false ;
	for (k=1;k<12;k++)
          if ((k != me.number)&&(k != tact->firstResponsiblePlayer))
            {
	      //if the player chooses a position far away from his own he checks a bigger area for other players
	      if (dist(me.pos, middlePositionsCutOff[i])>BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))
		//now the player is very far away	
		{  
		  xFront=KEEPOUT_X_FRONT_BIG;
		  xBack=KEEPOUT_X_BACK_BIG;
		  yFrontBack=KEEPOUT_Y_BIG;
		}
	      else
		//now we change the size of the keepoutarea continuously
		{
		  xFront=KEEPOUT_X_FRONT+dist(me.pos,middlePositionsCutOff[i])/(BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))*
		            (KEEPOUT_X_FRONT_BIG-KEEPOUT_X_FRONT);
		  xBack=KEEPOUT_X_BACK+dist(me.pos,middlePositionsCutOff[i])/(BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))*
			    (KEEPOUT_X_BACK_BIG-KEEPOUT_X_BACK);
		  yFrontBack=KEEPOUT_Y+dist(me.pos,middlePositionsCutOff[i])/(BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))*
		            (KEEPOUT_Y_BIG-KEEPOUT_Y);
		}
              if((world->ourTeam[k].pos.getX()<middlePositionsCutOff[i].getX()+xFront*(max.getX()-min.getX()))&&
                 (world->ourTeam[k].pos.getX()>middlePositionsCutOff[i].getX()-xBack*(max.getX()-min.getX()))&&
                 (world->ourTeam[k].pos.getY()<middlePositionsCutOff[i].getY()+yFrontBack*(max.getY()-min.getY()))&&
                 (world->ourTeam[k].pos.getY()>middlePositionsCutOff[i].getY()-yFrontBack*(max.getY()-min.getY()))&&
                 (world->ourTeam[k].age <MAXAGE))
		{
		  playerNear[i]=true;
                  //just for the traces
                  Vector2d KeepoutMax;
		  KeepoutMax.setX(middlePositionsCutOff[i].getX()+xFront*(max.getX()-min.getX()));
		  KeepoutMax.setY(middlePositionsCutOff[i].getY()+yFrontBack*(max.getY()-min.getY()));
		  Vector2d KeepoutMin;
                  KeepoutMin.setX(middlePositionsCutOff[i].getX()-xBack*(max.getX()-min.getX()));
		  KeepoutMin.setY(middlePositionsCutOff[i].getY()-yFrontBack*(max.getY()-min.getY()));
                  FREEPOSTRC(SBOX(KeepoutMin,KeepoutMax,MDot));

#ifdef FREEPOSTRACE
    {
      int t(0),x(0),y(0);		  
      ostrstream tmp;
      for (t=0; t<NumberofPointsX; t++)
        if((NumberofPointsY*t<i)&&(i<(t+1)*NumberofPointsY+1))
	  x=t ;
      y=i-(x*NumberofPointsY)-1;
      tmp << "gotoFreePos -> Player_Near Position (" <<  x  << "," << y <<")   Player " << k <<char(0);
      char * str= tmp.str();
      FREEPOSTRC(MLog(str));
      delete str;
    }
#endif

		}
	    }
      }

    //Downgrade positions too far away
    for (i=0; i<(NumberofPointsX*NumberofPointsY+1); i++)
      if (dist(middlePositionsCutOff[i], me.pos)> FAR_AWAY)
       {      
        middleGrades[i].downgrade();

#ifdef FREEPOSTRACE
    {
     int t(0),x(0),y(0);		  
     ostrstream tmp;
     for (t=0; t<NumberofPointsX; t++)
       if((NumberofPointsY*t<i)&&(i<(t+1)*NumberofPointsY+1))
	 x=t ;
     y=i-(x*NumberofPointsY)-1;
     tmp << "gotoFreePos ->  Position ( " << x  << ", " << y << ")  downgraded" << char(0);
     char * str= tmp.str();
     FREEPOSTRC(MLog(str));
     delete str;
    }
#endif 
       }
    

    BestGrade.set_IMPOSSIBLE();
    //Which position has best grade?
    mindistNumber=0;
    for(i=0;i<(NumberofPointsX*NumberofPointsY+1);i++)
      {  
	if ((BestGrade.is_worse_than(middleGrades[i]))&&(playerNear[i]==false)) 
	{
	  BestGradeNumber=i;
	  BestGrade=middleGrades[i];
	}
        else 
          //if two positions have the same grade take the nearest to real position
	  if  (BestGrade.is_like(middleGrades[i])) 
	    if  ((dist(me.pos,middlePositionsCutOff[i]) < dist(me.pos,middlePositionsCutOff[BestGradeNumber]))&&(playerNear[i]==false))
	      BestGradeNumber=i;      
	if ((dist(middlePositionsCutOff[i],LastBestPos)<dist(middlePositionsCutOff[mindistNumber],LastBestPos))&&(playerNear[i]==false))
	  mindistNumber=i;
      } 
  FREEPOSTRC(POSITION(middlePositionsCutOff[BestGradeNumber]));
#ifdef FREEPOSTRACE
 { 
   int t(0),x(0),y(0);		  
   ostrstream tmp;
   for (t=0; t<NumberofPointsX; t++)
     if((NumberofPointsY*t<BestGradeNumber)&&(BestGradeNumber<(t+1)*NumberofPointsY+1))
       x=t ;
   y=BestGradeNumber-(x*NumberofPointsY)-1;
   tmp << "gotoFreePos -> BestGrade  Position ( " << x  << ", " << y << ")  Grade" << BestGrade << char(0);
   char * str= tmp.str();
   FREEPOSTRC(MLog(str));
   delete str;
 }
#endif
  

  BestGrade.downgrade();
  //return the best position
  if (middleGrades[BestGradeNumber].is_better_than(Grade::POOR))
    if((BestGrade.is_better_than(middleGrades[mindistNumber]))||(LastBestPos.getX()==1000))
      {
	FREEPOSTRC(LINE(me.pos,middlePositionsCutOff[BestGradeNumber]));
	FREEPOSTRC(CIRCLE(middlePositionsCutOff[BestGradeNumber],3));
	FREEPOSTRC(MLog("free. pos-> going to position of best grade number"));
	// if we attack the offensive players should move forwards not backwards
        if((tact->ourTeamOffender(myTactNo))
	   &&(middlePositionsCutOff[BestGradeNumber].getX()<world->ourTeam[tact->firstResponsiblePlayer].pos.getX()-10 )
	   &&(world->ourTeam[tact->firstResponsiblePlayer].pos.getX()<(FIELD_LENGTH/2-3/4*PENALTY_AREA_LENGTH)) )
	      middlePositionsCutOff[BestGradeNumber].setX(max.getX());
	LastBestPos=middlePositionsCutOff[BestGradeNumber];
	if (dist(me.pos, middlePositionsCutOff[BestGradeNumber])>BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))
	  lookPos=middlePositionsCutOff[BestGradeNumber];
	return middlePositionsCutOff[BestGradeNumber];
      }
    else
      {
	FREEPOSTRC(LINE(me.pos,middlePositionsCutOff[mindistNumber]));
	FREEPOSTRC(CIRCLE(middlePositionsCutOff[mindistNumber],3));
	FREEPOSTRC(MLog("free. pos-> going to position of mindist number"));
	// if we attack the offensive players should move forwards not backwards
        if((tact->ourTeamOffender(myTactNo))
	   &&(middlePositionsCutOff[mindistNumber].getX()<world->ourTeam[tact->firstResponsiblePlayer].pos.getX()-10 )
	   &&(world->ourTeam[tact->firstResponsiblePlayer].pos.getX()<(FIELD_LENGTH/2-3/4*PENALTY_AREA_LENGTH)) )
	      middlePositionsCutOff[mindistNumber].setX(max.getX());
	LastBestPos=middlePositionsCutOff[mindistNumber];
	if (dist(me.pos, middlePositionsCutOff[mindistNumber])>BIGGER_KEEPOUT_AREA*(max.getY()-min.getY()))	
	  lookPos=middlePositionsCutOff[mindistNumber];
	return middlePositionsCutOff[mindistNumber];
      }
  else 
    {
      LastBestPos.setX(1000);
      LastBestPos.setY(1000);
      FREEPOSTRC(MLog("free. pos-> going to standard pos, best pass too bad"));
      return posGoal;
    };
  

};


Vector2d PositionModule::coverOpponent()
{
      	coveredOpponent = world->otherTeamClosestTo(me.pos);
	posGoal = CalCoverPos(coveredOpponent);
	goToPrecision = 2.0; //[C]   
	priority = 3;//[C]
	if (world->otherTeam[coveredOpponent].pos.getX() < -FIELD_LENGTH/4) priority -= 1;
	if (world->otherTeam[coveredOpponent].pos.getX() < -FIELD_LENGTH/2+PENALTY_AREA_LENGTH+7) priority -= 1;

#ifdef COVERTRACE
    {
	ostrstream tmp;
	tmp << "coverOpp-> Covering Opp Nr. " 
	    << coveredOpponent << char(0);
	char * str = tmp.str();
	COVERTRC(MLog(str));
	delete str;
    };
#endif

    return posGoal;
};

Vector2d PositionModule::CalCoverPos(int CoveredOpp){
  double coverDistance = 2.0; //[C]
  Vector2d reachBallPos = tact->otherTeamReachBallPos[tact->firstDangerousOpponent];
  Vector2d oppPos = world->otherTeam[CoveredOpp].pos;
  Vector2d coverVector =  reachBallPos - oppPos;

    //simple solution
    Vector2d goToVector = oppPos + (coverDistance / norm(coverVector)) * coverVector;
    
    //try to stay near home position
    if (norm(coverVector)>2*coverDistance){
	Vector2d homePos = positionEval->GetPlayerPosition(world,me.number);
	double minDist = dist(goToVector,homePos);
	Vector2d step = 0.1 * (norm(coverVector)-2*coverDistance)/norm(coverVector) * coverVector;
	    for (int i=1; i<10; i++){
		Vector2d test = oppPos + (coverDistance / norm(coverVector)) * coverVector 
		                       + i*step; 
		if (dist(homePos,test) < minDist){
		    minDist=dist(homePos,test);
		    goToVector = test;
		}
	    }
    }
    
#ifdef COVERTRACE
    Comm::sendMCmd(POSITION(goToVector));
#endif

    return goToVector;
};

void PositionModule::SupportPlayer()
{
  Vector2d suppPos=positionEval->GetPlayerPosition(world,SupportPlayerNo);

  if (world->ourTeam[SupportPlayerNo].age<5) 
    {
      Vector2d realPos=world->ourTeam[SupportPlayerNo].pos;
      double howfar=
	1-Step(3,dist(suppPos,positionEval->GetPlayerPosition(world,me.number)),
	       dist(suppPos,realPos));
      
#ifdef SUPPTRACE
      ostrstream tmp;
      tmp << "Supp. Player-> howfar= " << howfar << char(0);
      { char * str= tmp.str();
      SUPPTRC(MLog(str));
      delete str; }
#endif

      realPos=suppPos-positionEval->GetPlayerPosition(world,me.number);
      posGoal+=2/3.*howfar*realPos;
      SUPPTRC(LINE(posGoal,me.pos));
      SUPPTRC(LINE(suppPos,positionEval->GetPlayerPosition(world,me.number)));
      SUPPTRC(POSITION(posGoal));
      SUPPTRC(CIRCLE(suppPos,3));
      SUPPTRC(CIRCLE(posGoal,3));
      SUPPTRC(CIRCLE(positionEval->GetPlayerPosition(world,me.number),3));
    }
  else
    {
      SUPPTRC(MLog("supp-> resp. player age too old"));
    }
};
