//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


bool
StandardSituationModule::doingStdKickUs()
{
  Vector2d targetPos;
  Vector2d ball = world->ball.pos;

  // report tactical action
  tact->reportOurTacticalAction( TA_CONTROL, world->my.number );

  double stamina = world->my.staminaIncMax;
  // if we have enough stamina run faster
  if( world->my.stamina > activeStaminaFactor * world->my.staminaIncMax + EFFORT_DEC_THR * STAMINA_MAX )
    stamina = activeStaminaFactor * world->my.staminaIncMax;

  // reach a position not too close to the ball first...
  if( goToPhase <= 1 )
    {
      targetPos = ball + ( 2.0 * world->my.kickRange() ) * kickUsOffset;
      DECISION( "My targetPos (phase 1): " << targetPos );
      if( !goingTo( targetPos, stamina ) )
	goToPhase = 2;
    }

  // ...then approach ball
  if( goToPhase == 2 )
    {
      targetPos = ball + ( BALL_SIZE + world->my.size + 0.1 * world->my.kickableMargin ) * kickUsOffset;
      DECISION( "My targetPos (phase 2): " << targetPos );
      if( !goingTo( targetPos, stamina ) )
	{
	  bool waitForPasspartners = true;
	  int myNum = world->my.number;

	  for( int pnum = 1; pnum <= 11; pnum++ )
	    if( pnum != myNum &&
		world->ourTeam[pnum].age <= 2 &&
		passEval->ourEval( world, tact, myNum, pnum ).is_better_than( Grade::AVERAGE ) )
	      {
		waitForPasspartners = false;
		break;
	      }

	  // turn body to ball
	  AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );
	  if( tcmd.state != AdvSkills::RFINISHED )
	    skills->exec( tcmd );
	  else
	    skills->noAction();

	  waitingTime--;

	  if( waitingTime <= 0 ||
	      !waitForPasspartners ||
	      goalShotEval->ourEval( world, world->my.pos ).is_better_than( Grade::AVERAGE ) ||
	      timeLeft <= 0 )
	    {
	      stdSitFinished = true;
	      wantToTurnHead = true;
	      thwish.lookAtBodyDir( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	      DECISION( "Standardsituation finished" );
	    }
	  else
	    lookAround();
	}
    }

  return true;
}


bool
StandardSituationModule::doingKickOff()
{
  // move to standard kick off positions and turn around after that
  Vector2d moveTarget( -0.5 * FIELD_LENGTH + 5.0, 0.0 );   // position for the goalie
  if( world->my.number != world->ourGoalieNo )
    moveTarget = positionEval->GetPlayerPosition( world, world->my.number );

  int kickOffPlayer = ( world->ourGoalieNo == 10 ? 11 : 10 );

  if( world->ourTeam[world->my.number].tacticalNo == kickOffPlayer && weAreKickingOff )
    moveTarget.setXY( -BALL_SIZE - world->my.size, 0.0 );

  if( world->my.distTo( moveTarget ) > world->my.kickRange() ||
      ( world->ourTeam[world->my.number].tacticalNo == kickOffPlayer && 
	!world->my.ballKickable( world->ball ) ) )
    {
      if( world->playMode == PLAY_MODE_BEFORE_KICK_OFF ||
	  world->playMode == PLAY_MODE_GOAL_US ||
	  world->playMode == PLAY_MODE_GOAL_THEM )
	{
	  skills->basicMove( moveTarget.getX(),moveTarget.getY() );
	  DECISION( "doingKickOff: move player to kickoff position at " << moveTarget );
	  return true;
	}      
      DECISION( "doingKickOff: move not possible, walking to" << moveTarget );
      skills->exec( skills->goToPosition( moveTarget, world->my.kickRange() ) );
      return true;
    }

  lookAround();   // always turn head in this situation

  if( world->playMode == PLAY_MODE_BEFORE_KICK_OFF )
    {
      DECISION( "doingKickOff: turn the player (cause we are the MRB)" );
      skills->basicTurn( M_PI / 4.0 );
      thwish.lookAtBodyDir( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
      return true;
    }

  if( !weAreKickingOff || !world->my.ballKickable( world->ball ) )
    {
      DECISION( "doingKickOff: turn to ball" );
      AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );
      if( tcmd.state != AdvSkills::RFINISHED )
	skills->exec( tcmd );
      else
	skills->noAction();
      thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
      return true;
    }
   
  // I am responsible for the kick off

  // report tactical action
  tact->reportOurTacticalAction( TA_CONTROL, world->my.number );

  // turn to ball if needed
  AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );
  if( tcmd.state != AdvSkills::RFINISHED )
    {
      skills->exec( tcmd );
      return true;
    }

  // wait until playmode is KICK_OFF_US
  if( world->playMode != PLAY_MODE_KICK_OFF_US )
    {
      DECISION( "waiting for PLAY_MODE_KICK_OFF_US" );
      skills->noAction();
      return true;
    }

  stdSitFinished = true;
  DECISION( "Standardsituation finished" );
  skills->noAction();
  return true;
}


Grade
StandardSituationModule::evalRemainingKicksUs()
{
  Grade value( Grade::IMPOSSIBLE );

  // offset for kick in
  kickUsOffset = Vector2d( 0.0, ( world->ball.pos.getY() > 0.0 ? 1.0 : -1.0 ) );

  if( world->playMode == PLAY_MODE_KICK_IN_US )
    kickUsOffset = Vector2d( 0.0, ( world->ball.pos.getY() > 0.0 ? 1.0 : -1.0 ) );

  actPlayer = calcClosestDesiredPlayerTo( world->ball.pos + kickUsOffset );

  // maybe coach told us who is going to kick
  if( kickUsPlayer != UNKNOWN )
    actPlayer = kickUsPlayer;

  if( world->my.number == actPlayer )
    {
      value.set_ABSOLUTE();
      if( waitingTime < 0 )
	waitingTime = standardWaitingTime;
    }

  returnGRADE( value );
}


Grade
StandardSituationModule::evalFreeKickUs()
{
  Grade value( Grade::IMPOSSIBLE );

  // check if the goalie has catched the ball
  if( world->playMode == PLAY_MODE_GOALIE_CATCH_US ||
      ( world->ourGoalieNo != UNKNOWN &&
	world->ball.inOurPenaltyArea() &&
	world->ourTeam[world->ourGoalieNo].distTo( world->ball ) <
	world->ourTeam[world->ourGoalieNo].kickRange() ) )
    {
      DECISION( "goalie catched the ball " << world->cyclesSinceCatch << " cycles ago so goalieMove&Kick" );
      goalieDoesFreeKick = true;
      goalieMayMove = true;
    }
  else
    {
      DECISION( "goalie has not catched the ball so normal freekick" );
      goalieDoesFreeKick = false;
    }

  if( goalieDoesFreeKick )
    {
      if( iAmGoalie )
	{
	  value.set_ABSOLUTE();
	  if( waitingTime < 0 )
	    waitingTime = 2 * standardWaitingTime;   // [C]
	}
      else
	{
	  // get out of penalty area
	  if( world->my.inOurPenaltyArea() )
	    value.set_ABSOLUTE();
	}
      returnGRADE( value );
    }

  kickUsOffset = ( world->ball.pos.getY() > FIELD_WIDTH / 6.0 ? Vector2d( 0.0, 1.0 ) :
		   world->ball.pos.getY() < -FIELD_WIDTH / 6.0 ? Vector2d( 0.0, -1.0 ) :
		   Vector2d( -1.0, 0.0 ) );

  actPlayer = calcClosestDesiredPlayerTo( world->ball.pos + kickUsOffset );

  // maybe coach told us who is going to kick
  if( kickUsPlayer != UNKNOWN )
    actPlayer = kickUsPlayer;

  if( world->my.number == actPlayer )
    {
      value.set_ABSOLUTE();
      if( waitingTime < 0 )
	waitingTime = standardWaitingTime;
      if( goToPhase == 0 )
	goToPhase = 1;
    }

  returnGRADE( value );
}


bool 
StandardSituationModule::doingFieldPlayerGoalKickUs()
{
  // check if we have to leave the penalty area and do it
  bool getOutOfPenaltyArea = false;
  Vector2d dest = targetPos[world->my.tacticalNo]; 

  switch( tact->ourSystem.numDefenders )
    {
    case 4:
      if( world->my.tacticalNo != 2 && world->my.tacticalNo != 5 )
	getOutOfPenaltyArea = true;
      break;
    case 3:
      if( world->my.tacticalNo != 2 && world->my.tacticalNo != 4 )
	getOutOfPenaltyArea = true;
      break;
    default:
      if( world->my.tacticalNo != 2 && world->my.tacticalNo != 3 )
	getOutOfPenaltyArea = true;
      break;
    }

  if( getOutOfPenaltyArea )
    {
      dest = positionEval->GetPlayerPosition( world, world->my.number );
      if( inOurPenaltyArea( dest ) )
	dest.setX( -FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH + 8.0 );
      AdvSkills::TurtleCommand tcmd = skills->goToPosition( dest, 1 );
      if( tcmd.state != AdvSkills::TFINISHED )
	skills->exec( tcmd );
      else
	skills->noAction();
      lookAround();
      return true;
    }

  // if not yet reached go to wanted positon and wait for ball
  Vector2d leftPos( -FIELD_LENGTH / 2.0 + GOAL_AREA_LENGTH, -GOAL_AREA_WIDTH / 2.0 );
  Vector2d rightPos( -FIELD_LENGTH / 2.0 + GOAL_AREA_LENGTH, GOAL_AREA_WIDTH / 2.0 );

  bool goalieShot = ( sqrnorm( leftPos - world->ball.pos ) > 4.0 &&
		      sqrnorm( rightPos - world->ball.pos ) > 4.0 );

  if( !goalieShot )
  {
      if( goingTo( dest, STAMINA_INC_MAX ) )
      {
	  DECISION( "going to " << dest );
	  return true;
      }
      // ok, we have reached our position, now wait
      DECISION( "destination " << dest << " reached, now waiting" );
      skills->noAction();
      lookAround();
      return true;
  }

  // check if we need to intercept ball
  bool icpt = false;

  if( !world->my.ballKickable( world->ball ) )
    if( world->my.distTo( world->ball.pos ) < 1.5 * VISIBLE_DISTANCE )
      icpt = true;

  // check if I want to turn to the ball
  if( !icpt && !world->my.ballKickable( world->ball ) )
    {
      AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );

      if( tcmd.state != AdvSkills::RFINISHED )
	{
	  skills->exec( tcmd );
	  thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	  DECISION( "turning to ball" );
	  wantToTurnHead = true;
	  return true;
	}
    }

  // we need to intercept
  if( icpt )
    {
      WBall dummyBall;
      AdvSkills::TurtleCommand tcmd =
	skills->controlBall( world->my.stamina - EFFORT_DEC_THR * STAMINA_MAX,
			     skills->safeCircle, dummyBall );
      if( tcmd.state != AdvSkills::TFINISHED )
	{
	  skills->exec( tcmd );
	  DECISION( "controlling ball, meeting at " << dummyBall.pos );
	  thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  return true;
	}
    }

  // we intercepted, so go on...
  if( world->my.ballKickable( world->ball ) )
    {
      AdvSkills::TurnCommand tcmd;
      AdvSkills::KickCommand kcmd;

      DECISION( "ball in kickrange" );

      // first stop the ball
      kcmd = skills->stopBall();
      if( kcmd.state == AdvSkills::KSTOPPING && sqrnorm( world->ball.v ) > 0.05 )
	{
	  skills->exec( kcmd );
	  DECISION( "stopping ball" );
	  thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  return true;
	}

      // report tactical action
      tact->reportOurTacticalAction( TA_CONTROL, world->my.number );

      // now turn towards other goal
      tcmd = skills->turnTo( world->otherGoal );

      if( tcmd.state != AdvSkills::RFINISHED )
	{
	  skills->exec( tcmd );
	  DECISION("turning to other goal");
	  thwish.lookAtBall( Grade::ABSOLUTE, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  return true;
	}

      // now place ball in front of us
      if( fabs( ( world->ball.pos - world->my.pos ).getAngle() ) > M_PI_4 )
	{
	  DECISION( "placing ball" );
	  kcmd = skills->placeBall( Vector2d( skills->safeCircle, 0.0 ) );
	  skills->exec( kcmd );
	}
      else
	{
	  skills->noAction();
	}
      
      lookAround();
      waitingTime--;
      DECISION( "turned to ball, looking around" );

      if( waitingTime <= 0 || timeLeft <= 0 )
	{
	  stdSitFinished = true;
	  DECISION( "Standardsituation finished" );
	  return true;
	}

      return true;
    }

  DECISION( "Ball not kickable after trying to control it!!!" );
  skills->noAction();
  return false;
}
