//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef _TACTICALMODULE_H_
#define _TACTICALMODULE_H_


#include "worldmodel.h"
#include "AdvSkills.h"
#include "Config.h"
#include <iostream>
#include "grade.h"


///////////////////////////////////////////////////////////////////////////////////////////////
//
// enum TurnHeadWishState
//
// tells where a player wishes to look at
//
///////////////////////////////////////////////////////////////////////////////////////////////

enum TurnHeadWishState
{
  TH_UNKNOWN,
  TH_BALL,
  TH_OURTEAM,
  TH_OPPTEAM,
  TH_OURGOAL,
  TH_OPPGOAL,
  TH_POSITION,
  TH_BODYDIR,
  TH_LOOKAROUND,
  TH_LOOKAUDIO
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// class TurnHeadWish
//
// collects data about a player's turn head wish
//
///////////////////////////////////////////////////////////////////////////////////////////////

class TurnHeadWish
{
public:
  TurnHeadWish() { grade.set_IMPOSSIBLE(); type = TH_UNKNOWN; number = UNKNOWN; }
  ~TurnHeadWish() {}

  void lookNowhere( Grade _grade, view_width _vw ) { grade = _grade; type = TH_UNKNOWN; vWidth = _vw; }
  void lookAtBall( Grade _grade, view_width _vw ) { grade = _grade; type = TH_BALL; vWidth = _vw; }
  void lookAtOurPlayer( Grade _grade, int _number, view_width _vw )
  { grade = _grade; type = TH_OURTEAM; number = _number; vWidth = _vw; }
  void lookAtOppPlayer( Grade _grade, int _number, view_width _vw )
  { grade = _grade; type = TH_OPPTEAM; number = _number; vWidth = _vw; }
  void lookAtOurGoal( Grade _grade, view_width _vw ) { grade = _grade; type = TH_OURGOAL; vWidth = _vw; }
  void lookAtOppGoal( Grade _grade, view_width _vw ) { grade = _grade; type = TH_OPPGOAL; vWidth = _vw; }
  void lookAtPos(  Grade _grade, Vector2d _pos, view_width _vw )
  { grade = _grade; type = TH_POSITION; pos = _pos; vWidth = _vw; }
  void lookAtBodyDir( Grade _grade, view_width _vw ) { grade = _grade; type = TH_BODYDIR; vWidth = _vw; }
  void lookAround( Grade _grade, view_width _vw ) { grade = _grade; type = TH_LOOKAROUND; vWidth = _vw; }
  void lookAudio( Grade _grade, view_width _vw ) { grade = _grade; type = TH_LOOKAUDIO; vWidth = _vw; }

  friend std::ostream& operator << ( class std::ostream &os, const TurnHeadWish thw )
  {
    os << " wish to look at "
       << ( thw.type == TH_UNKNOWN ? "nothing" :
	    thw.type == TH_BALL ? "ball" :
	    thw.type == TH_OURTEAM ? "teammate" :
	    thw.type == TH_OPPTEAM ? "opponent" :
	    thw.type == TH_OURGOAL ? "our goal" :
	    thw.type == TH_OPPGOAL ? "opp goal" :
	    thw.type == TH_POSITION ? "position" :
	    thw.type == TH_BODYDIR ? "body dir" :
	    thw.type == TH_LOOKAROUND ? "landscape" : 
	    thw.type == TH_LOOKAUDIO ? "audio info" : "nirvana???" );
    if( thw.type == TH_OURTEAM || thw.type == TH_OPPTEAM )
      os << " " << thw.number;
    else if( thw.type == TH_POSITION )
      os << " " << thw.pos;
    os << ", view width: "
       << ( thw.vWidth == VIEW_WIDTH_WIDE ? "wide" :
	    thw.vWidth == VIEW_WIDTH_NORMAL ? "normal" :
	    thw.vWidth == VIEW_WIDTH_NARROW ? "narrow" : "undefined" );
    os << ", grade: " << thw.grade;
    return os;
  }

  Grade    grade;    // grade of this wish

private:

  friend class TacticalModule;

  TurnHeadWishState type;
  int               number;   // player number (in cases TH_OURTEAM and TH_OTHERTEAM)
  Vector2d          pos;      // position to look at
  view_width        vWidth;   // desired view width
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// enum BallControlState
//
// tells if (and how) a certain team (not player!) controls the ball
//
///////////////////////////////////////////////////////////////////////////////////////////////

enum BallControlState
{
  BC_SAFE,       // ball is under control
  BC_DANGER,     // ball is still under control, but not safe
  BC_LOST,       // opponent controls ball
  BC_UNKNOWN     // decision impossible
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// enum TacticalActionState
//
// tells what a certain team (not player!) is doing at the moment
//
///////////////////////////////////////////////////////////////////////////////////////////////

enum TacticalActionState
{
  TA_CONTROL,    // a team member controls the ball
  TA_PASS,       // team is playing a pass
  TA_GOALSHOT,   // a team member tried a goal kick
  TA_REACT,      // team tries to (re)conquer the ball
  TA_UNKNOWN     // decision impossible
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// struct PlayerPassInfo
//
// holds information about other players as pass partners (received via player communication)
//
///////////////////////////////////////////////////////////////////////////////////////////////

struct PlayerPassInfo
{
  Grade grade;
  int   xIndex;   // indices for rangepass grid
  int   yIndex;
  long  age;      // age of this information
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// enum TacticalActionQuality
//
// tells where a tactical action comes from
//
///////////////////////////////////////////////////////////////////////////////////////////////

enum TacticalActionQuality
{
  TQ_REPORTED,          // player reported that action (i.e. performed it himself)
  TQ_AUDIAL_REPORTED,   // a team member told us about his reported action
  TQ_AUDIAL,            // told by a team member but lesser quality
  TQ_GUESSED,           // guessed from worldmodel data
  TQ_UNKNOWN            // decision impossible
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// enum PlayerTypeState
//
// tells if some player can be of some type
//
///////////////////////////////////////////////////////////////////////////////////////////////

enum PlayerTypeState
{
  PTS_IMPOSSIBLE,   // player can not be of that type
  PTS_SURE,         // player must be of that type (because all other types are impossible)
  PTS_UNKNOWN
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// class PlayerTypeData
//
// collects status and score for some player and player type
//
///////////////////////////////////////////////////////////////////////////////////////////////

class PlayerTypeData
{
public:
  PlayerTypeData( PlayerTypeState _state = PTS_UNKNOWN, int _score = 0 )
    : state( _state ), score( _score ) {}
  ~PlayerTypeData() {}

  PlayerTypeState   state;
  int               score;
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class TacticalAction
//
// holds TacticalActionState for a team and (in case of PASS or CONTROL)
// the numbers of the involved players
//
///////////////////////////////////////////////////////////////////////////////////////////////

class TacticalAction
{
public:
  TacticalAction( TacticalActionState _state = TA_UNKNOWN,
		  int _perPlayer = UNKNOWN,
		  int _recPlayer = UNKNOWN,
		  long _age = ETERNAL )
  {
    state = _state;
    performer = ( state == TA_UNKNOWN ? UNKNOWN : _perPlayer );
    receiver = ( state == TA_PASS ? _recPlayer : UNKNOWN );
    age = _age;
  }

  ~TacticalAction() {}

  TacticalActionState state;

  TacticalActionQuality quality;   // information quality

  long age;   // number of cycles passed since this action started

  // if state is TA_CONTROL  : number of player who holds the ball
  // if state is TA_PASS     : number of player who sent the pass
  // if state is TA_GOALSHOT : number of player who tried the goal kick
  // if state is TA_REACT    : number of player who should intercept (firstResponsible)
  // if state is TA_UNKNOWN  : UNKNOWN ( == -1 )
  int performer;

  // if state is TA_PASS     : number of player who receives the pass
  // otherwise               : UNKNOWN ( == -1 )
  int receiver;
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class TacticalSystem
//
// holds the system we currently play
//
///////////////////////////////////////////////////////////////////////////////////////////////

class TacticalSystem
{
public:
  TacticalSystem( int _numDef = 4, int _numDefMid = 2, int _numOffMid = 2, int _numOff = 2 );

  ~TacticalSystem() {}

  int numDefenders;
  int numDefMidfield;
  int numOffMidfield;
  int numMidfield;
  int numOffenders;
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Class TacticalModule
//
///////////////////////////////////////////////////////////////////////////////////////////////

class TacticalModule
{
public:
  TacticalModule( Worldmodel * _world, AdvSkills * _skills );

  // CAUTION: I did NOT implement the following two lines (currently not needed):
  // TacticalModule( const TacticalModule& tm );
  // TacticalModule& TacticalModule::operator = ( const TacticalModule& tm );

  ~TacticalModule();

  // current state of ball control
  BallControlState   ourBallControl;

  // tactical action our team currently performs
  TacticalAction     ourAction;

  // tactical system our team currently plays
  TacticalSystem     ourSystem;

  // edges for changing ourBallControl: (depends on first responsible player's role)
  // ourBallControl changes to BC_SAFE if our intercept edge is > ourEdge
  int   ourEdgeDefender;
  int   ourEdgeDefMidfield;
  int   ourEdgeOffMidfield;
  int   ourEdgeOffender;
  int   ourEdgeGoalie;

  // ourBallControl changes to BC_LOST if opponent's intercept edge is > otherEdge
  int   otherEdgeDefender;
  int   otherEdgeDefMidfield;
  int   otherEdgeOffMidfield;
  int   otherEdgeOffender;
  int   otherEdgeGoalie;

  // returns if player with number "n" of our team is a defender/midfield/offender.
  // If player's number or tactical number is UNKNOWN, false is returned.
  // Goalie is neither defender nor midfield nor offender.
  bool  ourTeamDefender( int n ) const;
  bool  ourTeamDefMidfield( int n ) const;
  bool  ourTeamOffMidfield( int n ) const;
  bool  ourTeamMidfield( int n ) const;
  bool  ourTeamOffender( int n ) const;

  // the three players who will reach the ball first
  // this is just a guess because of lack of reliable stamina information
  // CAUTION 1: ball control module should not blindly use these numbers
  // CAUTION 2: Numbers can be UNKNOWN ( == -1 ) !
  int   firstResponsiblePlayer;
  int   secondResponsiblePlayer;
  int   thirdResponsiblePlayer;

  // most dangerous opponents' numbers
  // (i.e. opponents who will reach the ball first)
  // CAUTION 1: ball control module should not blindly use these numbers
  // CAUTION 2: Numbers can be UNKNOWN ( == -1 ) !
  int   firstDangerousOpponent;
  int   secondDangerousOpponent;
  int   thirdDangerousOpponent;

  // information how teammates grade themselves as passpartners
  PlayerPassInfo playerPassInfo[12];

  // get approximated number of steps players will need to reach the ball
  // value is ETERNAL if player is too far from the ball
  long ourTeamReachBallTime[12];

  // same for opponent team
  long otherTeamReachBallTime[12];

  // get approximated position where players will reach the ball
  // value is player's position if he is too far from the ball
  Vector2d ourTeamReachBallPos[12];

  // same for opponent team
  Vector2d otherTeamReachBallPos[12];
  
  // [MW] teammate to listen to
  int attentionTo;

  // no. of cycles passed since we received a message from this teammate
  long cyclesSinceLastMessage;

  // Report a tactical action of OUR team to the tactical module.
  // This can e.g. be used by a player who sent a pass or tried a goalkick.
  // It should only be used to report actions the player can be sure of because
  // he is involved. Everything else should be detected by the tactical module itself.
  // Returns false if information is invalid.
  bool  reportOurTacticalAction( TacticalActionState state,
				 int perPlayer = UNKNOWN,
				 int recPlayer = UNKNOWN );

  bool  reportOurTacticalAction( TacticalActionState state,
				 const WBall &_simulatedBall, 
				 int perPlayer = UNKNOWN,
				 int recPlayer = UNKNOWN);

  // for soccerserver 8+ communication
  void encodeAudio( char* msg );

  // feed audio information into module
  bool feedbackAudio( SensorData data );
  bool decodeAudio( const char *message, int sender, MessageType &msgType, 
		    bool &listenToMe, TacticalAction &audioTAction );

  // feed new information about world into module
  bool feedbackWorld();

  // after the modules have decided what to do: decide if to communicate (and what)
  MessageType determineMessageType( const CycleAction &action, Grade passGrade, WBall &_simulatedBall );

  // communicate after reportOurTacticalAction() ?
  bool communicateAction;
  
  // [MW] communicate simulated ball
  bool communicateSimAction;
  WBall simulatedBall;

  // holds information about opponent players - do not call this if you are not the coach
  int *otherPlayerType() { return playerType; };

  // calculate a turn head command from the turn head wishes of the other modules
  AdvSkills::TurnHeadCommand calcTurnHead( TurnHeadWish goalTHW, TurnHeadWish passTHW,
					   TurnHeadWish ballhTHW, TurnHeadWish posTHW,
					   TurnHeadWish stdTHW );

private:

  Worldmodel * world;
  AdvSkills * skills;

  // tactical action communicated by team member
  TacticalAction   audioTAction;

  // tactical action agent reported (is stored here for later use in determineOurTacticalAction())
  TacticalAction   reportedTAction;

  // fill in first/second/third... variables and RBTime and RBPos arrays
  void determineResponsibility();

  // determine ball control value for our team
  void determineBallControl();

  // determine tactical action for our team
  bool determineOurTacticalAction();

  // only the coach uses the following stuff (when determining opponent player types)

  WPlayer           oldPlayer[12];        // opponent player data from before
  long              oldTime;              // time step where oldPlayer array comes from
  int               oldNumSubs;           // number of opponent substitutions

  PlayerTypeData*   playerTypeData[12];   // type data for each player and type
  int               playerType[12];

  // determine player types of opponent players
  void determineOtherPlayerTypes();
};


// put opponent player types to AudioMsg-Stream
class AudioMsg& operator << ( AudioMsg &aumsg, int *playerType );

#endif  // _TACTICALMODULE_H_
